from unittest import TestCase

from . import interpret_strings_like


class InterpretTest(TestCase):
    def test_bool(self):
        """ Testing boolean interpretation"""
        m = interpret_strings_like
        abool = True
        self.assertEqual(True, m("True", abool))
        self.assertEqual(False, m("False", abool))
        self.assertEqual(True, m("True", abool))
        self.assertEqual(False, m("False", abool))
        self.assertEqual(True, m("1", abool))
        self.assertEqual(False, m("0", abool))
        self.assertRaises(ValueError, m, "", abool)
        self.assertRaises(ValueError, m, "a", abool)



_multiprocess_can_split_ = True  # Run parallel mcdp_lang_tests
from .compmake_test import *

# Load all mcdp_lang_tests (helps with nose multiprocess)
from . import test_blocked
from . import test_invalid_functions
from . import test_more
from . import test_priorities
from . import test_storage
from . import test_progress
from . import test_syntax
from . import test_unpickable_result



# from ..structures import CompmakeException, KeyNotFound, SerializationError
# from StringIO import StringIO
# import cPickle as pickle
# import sys
#
#
# class RedisInterface:
#     host = 'localhost'
#     port = 6379
#
#     # XXX Add dbname
#     local_cache = {}
#
#     @staticmethod
#     def __str__():
#         return "Redis backend"
#
#     @staticmethod
#     def supports_concurrency():
#         return True
#
#     # Storage public interface
#     @staticmethod
#     def get(name):
#         if name in RedisInterface.local_cache:
#             sys.stderr.write('+')
#             return RedisInterface.local_cache[name]
#
#         k = key2rediskey(name)
#         s = get_redis().get(k)
#
#         if s is None:
#             raise KeyNotFound('Key %s does not exist anymore' % name)
#
#         if not isinstance(s, str):
#             raise CompmakeException('I usually put string-string values in\
# the db, however I found %s (%s). Key is %s' % (s, type(s), k))
#         try:
#             value = string2object(s)
#         except Exception as e:
#             tmp_core = '/tmp/pickle_core'
#             open(tmp_core, 'w').write(s)
#             msg = ("Could not load cache %s.\n Dumped file %s.\n Error: '%s'" %
#                 (name, tmp_core, e))
#             print(msg)
#             raise e
#
#         # Useful to get a sense what it's doing
#         # sys.stderr.write('Load %s [%.2fK]\n' % (name, len(s)/1000.0 ) )
#         return value
#
#     @staticmethod
#     def delete(name):
#         k = key2rediskey(name)
#         get_redis().delete(k)
#
#     @staticmethod
#     def exists(name):
#         k = key2rediskey(name)
#         return get_redis().exists(k)
#
#     @staticmethod
#     def set(name, value, precious=False):  # @ReservedAssignment
#         if not isinstance(name, str):
#             raise CompmakeException(
#                 'Panic: received %s (%s) as a key. I want strings.' %
#                 (name, type(name)))
#         k = key2rediskey(name)
#         s = object2string(value)
#
#         # Useful to get a sense what it's doing
#         # sys.stderr.write('Save %s [%.2fK]\n' % (name, len(s)/1000.0 ) )
#
#         assert(isinstance(k, str))
#         assert(isinstance(s, str))
#         get_redis().set(k, s)
#
#         if precious:
#             RedisInterface.local_cache[name] = value
#
#     @staticmethod
#     def keys(pattern='*'):
#         K = get_redis().keys(pattern=key2rediskey(pattern))
#         assert isinstance(K, str), \
#             'I think you have the wrong version of pyredis '
#         K = K.split()
#         assert(isinstance(K, list))
#         return [rediskey2key(k) for k in K]
#
#     @staticmethod
#     def reopen_after_fork():
#         global redis
#         redis = None
#         get_redis(force=True)
#
#     @staticmethod
#     def events_push(event):
#         ''' Pushes an event over the queue '''
#         r = get_redis()
#         r.rpush('compmake-events', object2string(event))
#
#     @staticmethod
#     def events_read():
#         ''' Returns a list of events '''
#         events = []
#         r = get_redis()
#         while True:
#             event_string = r.lpop('compmake-events')
#             if not event_string:
#                 break
#             events.append(string2object(event_string))
#         return events
#
# # Other utilities
#
#
# # Serialization device
# def object2string(obj):
#     sio = StringIO()
#     try:
#         pickle.dump(obj, sio, pickle.HIGHEST_PROTOCOL)
#     except Exception as e:
#         raise SerializationError('Cannot pickle object '
#                     'of class %s: %s' % (obj.__class__.__name__, e))
#     return sio.getvalue()
#
#
# def string2object(s):
#     sio = StringIO(s)
#     return pickle.load(sio)
#
#
# def key2rediskey(s):
#     return "compmake:%s" % s
#
#
# def rediskey2key(key):
#     return key.replace('compmake:', '', 1)
#
# redis = None
#
#
# def get_redis(force=False):
#
#     from redis import Redis #@UnresolvedImport
#     from redis.exceptions import ConnectionError #@UnresolvedImport
#     global redis
#     if redis is None or force:
#         #sys.stderr.write("Opening connection to Redis (host=%s)... " %
#         #                        RedisInterface.host)
#         try:
#             redis = Redis(host=RedisInterface.host, port=RedisInterface.port)
#         except ConnectionError as e:
#             raise CompmakeException(str(e))
#         #sys.stderr.write("done.\n")
#     return redis



import datetime


__all__ = [
    "isodate_with_secs",
]


def isodate_with_secs():
    """ E.g., '2011-10-06-22:54:33' """
    now = datetime.datetime.now()
    date = now.isoformat("-")[:19]
    return date



__all__ = [
    "CompmakeConstants",
    "DefaultsToConfig",
]


class CompmakeConstants:
    """ Arbitrary constants used in the code. """

    # DO NOT change these -- they are part of Compmake's interface
    job_id_key = "job_id"
    extra_dep_key = "extra_dep"
    command_name_key = "command_name"

    # Compmake returns:
    # 0                      if everything all right
    #  RET_CODE_JOB_FAILED    if some job failed
    #  RET_CODE_COMPMAKE_BUG  if compmake itself had some errors
    RET_CODE_JOB_FAILED = 113
    RET_CODE_COMMAND_FAILED = 1
    RET_CODE_COMPMAKE_BUG = 114

    # Statuses ------------------------------------------------
    compmake_status_interactive = "interactive"
    # If run as a ssh-spawned slave session.
    # - Jobs cannot be created
    compmake_status_slave = "slave"
    # If run embedded in the user program, when executed by python
    compmake_status_embedded = "embedded"

    # debug_origin_of_prints = True
    debug_check_invariants = False  # TODO: make config

    disable_interproc_queue = False

    extra_checks_job_states = False

    tolerate_db_inconsistencies = True

    # Try to recover from anomalous situations
    try_recover = False

    aliases = {}


class DefaultsToConfig:
    """ Used to mean the param's default is from a config switch. """

    def __init__(self, switch: str):
        self.switch = switch



# -*- coding: utf-8 -*-

from .manager_multiprocessing import MultiprocessingManager
from compmake.events import publish
from compmake.jobs.queries import top_targets
from compmake.ui import COMMANDS_ADVANCED, ui_command
from compmake.ui.commands import raise_error_if_manager_failed
from compmake.constants import DefaultsToConfig

__all__ = [
    "parmake_pool",
]


@ui_command(section=COMMANDS_ADVANCED, dbchange=True)
def parmake_pool(job_list, context, cq, n=DefaultsToConfig("max_parallel_jobs"), recurse=False):
    """
        Parallel equivalent of "make", using multiprocessing.Pool. (buggy)

        Usage:

           parmake [n=<num>] [joblist]

     """

    publish(context, "parmake-status", status="Obtaining job list")
    job_list = list(job_list)

    db = context.get_compmake_db()
    if not job_list:
        job_list = list(top_targets(db=db))

    publish(context, "parmake-status", status="Starting multiprocessing manager (forking)")
    manager = MultiprocessingManager(num_processes=n, cq=cq, context=context, recurse=recurse)

    publish(context, "parmake-status", status="Adding %d targets." % len(job_list))

    manager.add_targets(job_list)

    publish(context, "parmake-status", status="Processing")
    manager.process()
    return raise_error_if_manager_failed(manager)



from . import CompmakeTest
from ..jobs import direct_children, direct_parents, make
from ..exceptions import UserError
from nose.tools import istest


def f1(*arg, **kwargs):
    """ Generic function """
    pass


def f2(*arg, **kwargs):
    """ Generic function """
    pass


def failing():
    """ A function that raises an exception """
    raise TypeError()


def uses_id(a, b, job_id):
    """ A function with a job_id arguement """
    pass


@istest
class Test1(CompmakeTest):
    def mySetUp(self):
        pass

    def testAdding(self):
        self.comp(f1)
        self.assertTrue(True)

    def testID(self):
        """ Check that the job id is correctly parsed """
        job_id = "terminus"
        c = self.comp(f1, job_id=job_id)
        self.assertEqual(c.job_id, job_id)
        make(job_id, context=self.cc)
        self.assertTrue(True)

    def testID2(self):
        """ Make sure we set up a warning if the job_id key
            is already used """
        self.assertTrue(self.comp(f1, job_id="ciao"))
        self.assertRaises(UserError, self.comp, f1, job_id="ciao")

    def testDep(self):
        """ Testing advanced dependencies discovery """
        cf1 = self.comp(f1)
        cf2 = self.comp(f2, cf1)
        self.assertTrue(cf1.job_id in direct_children(cf2.job_id, db=self.db))
        self.assertTrue(cf2.job_id in direct_parents(cf1.job_id, db=self.db))

    def testDep2(self):
        """ Testing advanced dependencies discovery (double) """
        cf1 = self.comp(f1)
        cf2 = self.comp(f2, cf1, cf1)
        self.assertTrue(cf1.job_id in direct_children(cf2.job_id, db=self.db))
        self.assertEqual(1, len(direct_children(cf2.job_id, db=self.db)))
        self.assertEqual(1, len(direct_parents(cf1.job_id, db=self.db)))

    def testDep3(self):
        """ Testing advanced dependencies discovery in dicts"""
        cf1 = self.comp(f1)
        cf2 = self.comp(f2, [1, {"ciao": cf1}])
        self.assertTrue(cf1.job_id in direct_children(cf2.job_id, db=self.db))
        self.assertTrue(cf2.job_id in direct_parents(cf1.job_id, db=self.db))

    def testJOBparam(self):
        """ We should issue a warning if job_id is used
            as a parameter in the function """
        self.comp(uses_id)
        self.assertRaises(UserError, self.comp, uses_id, job_id="myjobid")



import time


__all__ = [
    "EventSpec",
    "Event",
]


class EventSpec:
    """ This is a specification of the events that can be generated """

    def __init__(self, name, attrs=None, desc=None):
        if attrs is None:
            attrs = []
        self.name = name
        self.attrs = attrs
        self.desc = desc


class Event:
    """ This, instead, is an event itself """

    def __init__(self, name, **kwargs):
        self.name = name
        self.__dict__.update(kwargs)
        self.kwargs = kwargs
        self.timestamp = time.time()

    def __str__(self):
        return f"Event({self.name}, {self.kwargs})"



from compmake.context import Context
from compmake.storage.filesystem import StorageFilesystem
from compmake.unittests.compmake_test import CompmakeTest
from nose.tools import istest
from compmake.jobs.queries import definition_closure


def g2():
    pass


def gd(context):
    context.comp(g2)


def fd(context):
    context.comp_dynamic(gd)


def i2():
    pass


def id(context):  # @ReservedAssignment
    context.comp(i2)


def hd(context):
    context.comp_dynamic(id)


def mockup5(context, both):
    context.comp_dynamic(fd)
    if both:
        context.comp_dynamic(hd)


@istest
class TestDynamic5(CompmakeTest):
    def test_dynamic5(self):

        # first define with job and run
        mockup5(self.cc, both=True)
        self.assert_cmd_success("make recurse=1")

        self.assertJobsEqual("all", ["fd", "fd-gd", "fd-gd-g2", "hd", "hd-id", "hd-id-i2"])
        self.assertJobsEqual("done", ["fd", "fd-gd", "fd-gd-g2", "hd", "hd-id", "hd-id-i2"])

        self.assert_cmd_success("details hd-id")
        self.assert_cmd_success("details hd-id-i2")
        self.assertEqualSet(definition_closure(["hd-id"], self.db), ["hd-id-i2"])
        self.assertEqualSet(definition_closure(["hd"], self.db), ["hd-id", "hd-id-i2"])
        # now redo it
        self.db = StorageFilesystem(self.root, compress=True)
        self.cc = Context(db=self.db)

        mockup5(self.cc, both=False)
        self.assert_cmd_success("clean")
        self.assert_cmd_success("make recurse=1")
        self.assertJobsEqual("all", ["fd", "fd-gd", "fd-gd-g2"])
        self.assertJobsEqual("done", ["fd", "fd-gd", "fd-gd-g2"])



import sys

from nose.tools import istest

from . import CompmakeTest
from ..exceptions import CompmakeSyntaxError, UserError
from ..jobs import get_job_cache, set_job_cache
from ..structures import Cache, CMJobID
from ..ui import parse_job_list


def dummy():
    pass


@istest
class Test1(CompmakeTest):
    def mySetUp(self):
        # Removed when refactoring
        # remove_all_jobs()
        # reset_jobs_definition_set()

        self.jobs = [
            ("a", Cache.DONE),
            ("b", Cache.FAILED),
            ("c", Cache.NOT_STARTED),
            ("d", Cache.DONE),
            ("e", Cache.DONE),
            #             ('f', Cache.IN_PROGRESS),
            ("g", Cache.DONE),
            ("h", Cache.FAILED),
            ("i", Cache.DONE),
            ("ii", Cache.DONE),
            ("v_rangefinder_nonunif-plot_tensors_tex-0", Cache.DONE),
        ]

        for job_id, state in self.jobs:
            self.comp(dummy, job_id=job_id)
            cache = get_job_cache(job_id, db=self.db)
            cache.state = state
            set_job_cache(CMJobID(job_id), cache, db=self.db)

        self.all = set([job_id for job_id, state in self.jobs])
        selectf = lambda S: set([nid for nid, state_ in self.jobs if state_ == S])
        self.failed = selectf(Cache.FAILED)
        self.done = selectf(Cache.DONE)
        #         self.in_progress = selectf(Cache.IN_PROGRESS)
        self.not_started = selectf(Cache.NOT_STARTED)

    def selection(self, crit):
        return set([nid for nid, state in self.jobs if crit(nid, state)])

    def expandsTo(self, A, B):
        """ A, B can be:
        - set or list: list of jobs
        - string: passed to expands_jobs
        - lambda: passed to selection()
        """

        def expand_to_set(X):
            if isinstance(X, set):
                return X
            elif isinstance(X, list):
                return set(X)
            elif isinstance(X, type(lambda: 0)):
                return self.selection(X)
            elif isinstance(X, str):
                return set(parse_job_list(X, context=self.cc))
            else:
                assert False, "Wrong type %s" % type(X)

        a = expand_to_set(A)
        b = expand_to_set(B)

        try:
            self.assertEqualSet(a, b)
        except:
            sys.stdout.write("Comparing:\n\t- %s\n\t   -> %s \n\t- %s\n\t   -> %s. \n" % (A, a, B, b))
            raise

    def syntaxError(self, s):
        def f(x):  # it's a generator, you should try to read it
            return list(parse_job_list(x, context=self.cc))

        self.assertRaises(CompmakeSyntaxError, f, s)

    def userError(self, s):
        self.assertRaises(UserError, parse_job_list, s)

    def testCatchErrors(self):
        self.syntaxError("not")
        self.syntaxError("all not")
        self.syntaxError("all not")
        self.syntaxError("all in")
        self.syntaxError("in $all")
        self.syntaxError("all not e")

    def testSpecial(self):
        """ Test that the special variables work"""
        self.expandsTo("  ", set())
        self.expandsTo("all", self.all)
        self.expandsTo("failed", self.failed)
        self.expandsTo("done", self.done)
        self.expandsTo("DONE", self.done)

    #         self.expandsTo('in_progress', self.in_progress)

    def testBasicUnion(self):
        """ Testing basic union operator """
        self.expandsTo("failed e", self.failed.union("e"))
        self.expandsTo("e failed", self.failed.union("e"))

    def testNot(self):
        all_not_e = self.selection(lambda job, _: job != "e")
        self.expandsTo("e", ["e"])
        self.expandsTo("e*", ["e"])
        self.expandsTo("not e", all_not_e)
        self.expandsTo("not e*", all_not_e)
        self.expandsTo("all except e", all_not_e)
        self.expandsTo("not not e", ["e"])
        self.expandsTo("not not all", "all")
        self.expandsTo("not all", [])
        self.expandsTo("not all except all", [])
        self.expandsTo("not e except not e", [])
        self.expandsTo("not a b c except not a b c", [])
        self.expandsTo("not c except a ", "not a c")
        self.expandsTo("a in c  ", [])
        self.expandsTo("a in all  ", "a")
        self.expandsTo("all in all  ", "all")

    def testFailed(self):
        self.expandsTo("all except failed", lambda _, state: state != Cache.FAILED)
        self.expandsTo("not failed", lambda _, state: state != Cache.FAILED)

    def testIntersection(self):
        self.expandsTo("a b in a b c", ["a", "b"])
        self.expandsTo("a b c in d e", [])

    def tearDown(self):
        pass



import sys
import types
from collections import namedtuple
from typing import List, Optional, Union

from .visualization import compmake_colored
from ..exceptions import UserError
from ..utils import docstring_components, docstring_trim

# Storage for the commands
Command = namedtuple("Command", "function name doc alias section dbchange")
# Storage for the sections
Section = namedtuple("Section", "name desc order commands experimental")

# noinspection PyClassHasNoInit
class UIState:
    # name (string) -> tuple (function, name, docs, alias, section)
    commands = {}
    # holds alias -> name  (string->string) table
    alias2name = {}

    # section name
    # section name -> Section
    sections = {}

    last_section_name = None  # XXX


# ############ Definition of UI sections ##############


def ui_section(section_name, desc=None, order=None, experimental=False):
    if not section_name in UIState.sections:
        UIState.sections[section_name] = Section(
            name=section_name, desc=desc, order=order, commands=[], experimental=experimental
        )
    else:
        assert not desc and not order, "Description already given for section %s" % section_name

    UIState.last_section_name = section_name


GENERAL = "General commands"
VISUALIZATION = "Visualization"
ACTIONS = "Commands for making and cleaning jobs"
COMMANDS_ADVANCED = "Advanced commands and diagnostics"
# COMMANDS_CLUSTER = '(Experimental) Cluster commands'

ui_section(GENERAL, order=0)
ui_section(VISUALIZATION, order=1)
ui_section(ACTIONS, order=2)
# ui_section(COMMANDS_CLUSTER, order=2.5,
# desc='Experimental: These assume that you have a cluster '
# ' configuration file as explained in the documentation.',
# experimental=True)
ui_section(
    COMMANDS_ADVANCED, order=4, desc="These are advanced commands not for general use.", experimental=True
)


############# Helpers for defining commands ##############


# This is a decorator with arguments --
# see http://www.artima.com/weblogs/viewpost.jsp?thread=240845
# for an explanation. Also see for additional trick


def wrap(func, name, alias, section, dbchange):
    """ Decorator for a UI command -- wrapper for register_command """
    if name is None:
        name = func.__name__
    docs = func.__doc__
    register_command(name=name, func=func, docs=docs, alias=alias, section=section, dbchange=dbchange)
    return func


# @contract(alias="None|unicode|list(unicode)")
def ui_command(name=None, alias: Optional[Union[str, List[str]]] = None, section=None, dbchange=False):
    if alias is None:
        alias = []
    if isinstance(name, types.FunctionType):
        func = name
        return wrap(func, name=None, alias=[], section=None, dbchange=False)

    return lambda x: wrap(x, name, alias, section, dbchange)


def register_command(name, func, docs, alias=None, section=None, dbchange=False):
    if alias is None:
        alias = []
    if isinstance(alias, str):
        alias = [alias]
    if not section:
        section = UIState.last_section_name
    assert not name in UIState.commands, "Command %r already defined " % name
    assert docs is not None, "Command %r need docs." % name
    UIState.commands[name] = Command(
        function=func, name=name, doc=docs, alias=alias, section=section, dbchange=dbchange
    )
    assert section in UIState.sections, "Section %r not defined" % section
    UIState.sections[section].commands.append(name)
    for a in alias:
        assert not a in UIState.alias2name, 'Alias "%s" already used' % a
        assert not a in UIState.commands, 'Alias "%s" is already a command' % a
        UIState.alias2name[a] = name


def get_commands():
    return UIState.commands


# noinspection PyShadowingBuiltins
@ui_command(section=GENERAL)
def help(args):  # @ReservedAssignment
    """
        Prints help about the other commands. (try 'help help')

        Usage:

        @: help [command]

        If command is given, extended help is printed about it.
    """
    commands = get_commands()
    if not args:
        list_commands_with_sections()
    else:
        if len(args) > 1:
            msg = 'The "help" command expects at most one parameter.' " (got: %s)" % args
            raise UserError(msg)

        c = args[0]
        if not c in commands.keys():
            raise UserError("Command %r not found." % c)

        cmd = commands[c]
        # dbchange = cmd.dbchange
        s = "Command '%s'" % cmd.name
        s = s + "\n" + "-" * len(s)
        print(s)
        doc = docstring_trim(cmd.doc)
        print(doc)


def list_commands_with_sections(file=sys.stdout):  # @ReservedAssignment
    ordered_sections = sorted(UIState.sections.values(), key=lambda _section: _section.order)

    max_len = 1 + max([len(cmd.name) for cmd in UIState.commands.values()])
    for section in ordered_sections:
        is_experimental = section.experimental
        h = section.name
        if not is_experimental:
            h = compmake_colored(h, attrs=["bold"])
        h = h + " " + "-" * (79 - len(h))
        file.write("  ---- %s \n" % h)
        if section.desc:
            # XXX  multiline
            file.write("  | %s \n" % section.desc)
        for name in section.commands:
            cmd = UIState.commands[name]
            dbchange = cmd.dbchange

            dc = docstring_components(cmd.doc)
            short_doc = dc["first"]
            #             short_doc = cmd.doc.split('\n')[0].strip()
            if False:  # display * next to jobs affecting the DB
                if dbchange:
                    name += "*"
            n = name.ljust(max_len)
            if not is_experimental:
                n = compmake_colored(n, attrs=["bold"])
            file.write("  | %s  %s\n" % (n, short_doc))



# -*- coding: utf-8 -*-

import os
import shutil

from compmake.state import get_compmake_config
from compmake.ui.visualization import info
from zuper_commons.fs import friendly_path

__all__ = ["sync_data_up", "sync_data_down"]

up_arrow = "\u25B2"
down_arrow = "\u25BC"


# @contract(returns="list(unicode)")
def sync_data_up(context, skipsync=False):
    """ Synchronizes the data to the cloud.

        Returns the list of volume names.
    """
    syncdirs = get_compmake_config("multyvac_sync_up")
    if not syncdirs:
        return []

    volumes = set()
    for syncdir in syncdirs.split(":"):
        if not syncdir:
            continue
        v = sync_data_up_dir(syncdir, skipsync)
        volumes.add(v)

    return sorted(volumes)


def sync_data_up_dir(syncdir, skipsync=False):
    if not os.path.exists(syncdir):
        msg = "Dir %r does not exist." % syncdir
        raise ValueError(msg)

    syncdir = os.path.realpath(syncdir)

    if not syncdir or not syncdir[0] == "/":
        msg = "I expect the multyvac_sync_up dir to be an absolute path,"
        msg += " got %r." % syncdir
        raise ValueError(msg)

    vol, rest, rest_minus = get_volume_for_dir(syncdir)

    vol.mkdir(rest)
    if not skipsync:
        info(up_arrow + " Synchronizing directory %r up." % friendly_path(syncdir))
        vol.sync_up(syncdir, rest_minus)

    return vol.name


def get_volume_for_dir(dirname):
    if not dirname or dirname[0] != "/":
        msg = "Expected absolute path for %r." % dirname
        raise ValueError(msg)

    import multyvac

    # local='/data/work/datasets/00-ldr21-logs/'
    # Let's use the first path
    tokens = dirname.split("/")
    if tokens < 3:
        msg = "Got %r." % str(tokens)
        raise ValueError(msg)

    volume_name = tokens[1]
    mount_path = "/" + volume_name
    rest = "/".join(tokens[2:])

    volumes = [v.name for v in multyvac.volume.list()]  # @UndefinedVariable

    # print('obtained list of volumes %r' % volumes)
    if not volume_name in volumes:
        # print('creating new volume')
        multyvac.volume.create(volume_name, mount_path)  # @UndefinedVariable

    vol = multyvac.volume.get(volume_name)  # @UndefinedVariable

    rest_minus = "/".join(tokens[2:-1])

    return vol, rest, rest_minus


def get_sync_dirs_down():
    syncdirs = get_compmake_config("multyvac_sync_down")
    for syncdir in syncdirs.split(":"):
        if not syncdir:
            continue
        yield syncdir


def sync_data_down(context):
    """ Synchronizes the data from the cloud.

        Returns the list of volume names.
    """
    for syncdir in get_sync_dirs_down():
        sync_data_down_dir(syncdir)


def clean_cloud_out():
    """ Deletes all the content of the output directories """
    for syncdir in get_sync_dirs_down():
        clean_cloud_out_dir(syncdir)


def clean_cloud_out_dir(d):
    d = os.path.realpath(d)
    vol, _, _ = get_volume_for_dir(d)
    import multyvac

    multyvac_job_id = multyvac.submit(
        clean_cloud_out_dir_job, d, _vol=[vol.name], _name="Cleaning directory %r" % d
    )
    multyvac_job = multyvac.get(multyvac_job_id)
    multyvac_job.get_result()


def clean_cloud_out_dir_job(d):
    if os.path.exists(d):
        shutil.rmtree(d)
    os.makedirs(d)


def sync_data_down_dir(syncdir):

    if not os.path.exists(syncdir):
        os.makedirs(syncdir)

    syncdir = os.path.realpath(syncdir)

    if not syncdir or not syncdir[0] == "/":
        msg = "I expect the multyvac_sync_down dir to be an absolute path,"
        msg += " got %r." % syncdir
        raise ValueError(msg)

    vol, rest, rest_minus = get_volume_for_dir(syncdir)
    info(down_arrow + " Synchronizing directory %r down." % friendly_path(syncdir))
    vol.sync_down(rest, os.path.dirname(syncdir))



import argparse
import os
import subprocess
import sys
import traceback
from dataclasses import dataclass
from datetime import datetime
from subprocess import Popen
from typing import cast, Dict, List, NewType, Optional, Tuple, Union

from compmake.ui import compmake_console_gui
from networkx import descendants, DiGraph
from zuper_commons.fs import AbsDirPath, AbsFilePath, DirPath, read_ustring_from_utf8_file
from zuper_commons.text import get_md5
from zuper_commons.types import ZException, ZKeyError, ZValueError

from . import logger, Promise

TargetName = NewType("TargetName", str)


@dataclass
class RegularCommand:
    line: str


@dataclass
class MakeC:
    where: AbsDirPath
    filename: AbsFilePath
    targets: List[TargetName]


Command = Union[RegularCommand, MakeC]


@dataclass
class TargetInfo:
    dependencies: List[TargetName]
    commands: List[Command]


@dataclass
class MakefileParsed:
    C: AbsDirPath
    fn: AbsFilePath
    assignments: Dict[str, str]
    conditional_assignments: Dict[str, str]
    templates: Dict[TargetName, TargetInfo]
    targets: Dict[TargetName, TargetInfo]

    def get_target(self, k: TargetName) -> "MakeTarget":
        v = self.resolve_target_info(k)
        dependencies: List[Tuple[AbsDirPath, AbsFilePath, TargetName]]
        dependencies = []
        for _ in v.dependencies:
            key = (self.C, self.fn, _)
            dependencies.append(key)

        if len(v.commands) == 0:
            other_commands = False
        elif len(v.commands) == 1:
            c0 = v.commands[0]
            if isinstance(c0, MakeC):
                # go1(c0.where, c0.filename, bs=bs, only=c0.targets, extra_dependencies=dependencies)
                for d in c0.targets:
                    dependencies.append((c0.where, c0.filename, d))
                other_commands = False
            else:
                other_commands = True
        else:
            other_commands = True

        # OK this is not the best idea because
        # then when we say "pretend" for the sink, we don't update
        # anything else.
        # if not mtargets:  # only the sink depend externally
        # dependencies.extend(extra_dependencies)
        mt = MakeTarget(
            cwdir=self.C, filename=self.fn, target=k, dependencies=dependencies, other_commands=other_commands
        )
        return mt

    def resolve_target_info(self, k: TargetName) -> TargetInfo:
        if k in self.targets:
            return self.targets[k]
        for t, ti in self.templates.items():
            piece = match_template(t, k)
            if piece is None:
                # logger.info('no match', t=t, k=k)
                continue
            res = replace_pattern(ti, piece)
            # logger.debug(k=k, ti=res)
            return res

        msg = f"Cannot resolve {k!r}"
        raise ZKeyError(msg, mp=self)


def replace_command(c: Command, piece: str) -> Command:
    if isinstance(c, RegularCommand):
        line = c.line.replace("$*", piece)
        return RegularCommand(line)
    elif isinstance(c, MakeC):
        targets = [_.replace("$*", piece) for _ in c.targets]
        return MakeC(targets=cast(List[TargetName], targets), where=c.where, filename=c.filename)
    else:
        raise ZValueError(c=c)


def replace_pattern(ti: TargetInfo, piece: str) -> TargetInfo:
    dependencies = [_.replace("%", piece) for _ in ti.dependencies]
    commands = [replace_command(c, piece) for c in ti.commands]
    return TargetInfo(dependencies=cast(List[TargetName], dependencies), commands=commands)


import re


def match_template(template: str, k: str) -> Optional[str]:
    reg = template.replace("%", "(.*)")
    m = re.match(reg, k)
    if m is not None:
        return m.group(1)
    return None


@dataclass
class MakeTarget:
    cwdir: AbsDirPath
    filename: AbsFilePath
    target: TargetName
    other_commands: bool
    dependencies: List[Tuple[AbsDirPath, AbsFilePath, TargetName]]

    def __post_init__(self):
        for _ in self.dependencies:
            if _ == (self.cwdir, self.filename, self.target):
                raise ZValueError("circular", _=self)

    def get_local_deps(self) -> List[TargetName]:
        targets = [c for a, b, c in self.dependencies if a == self.cwdir and b == self.filename]
        return targets

    def get_local_deps2(self) -> List[Tuple[AbsDirPath, AbsFilePath, TargetName]]:
        targets = [(a, b, c) for a, b, c in self.dependencies if a == self.cwdir and b == self.filename]
        return targets


@dataclass
class BuildSystem:
    _files: Dict[Tuple[AbsDirPath, AbsFilePath], MakefileParsed]
    _targets: Dict[Tuple[AbsDirPath, AbsFilePath, TargetName], MakeTarget]

    def get_make_target(self, C: AbsDirPath, fn: AbsFilePath, k: TargetName) -> MakeTarget:
        key = (C, fn, k)
        if key not in self._targets:
            mp = self.get_mp(C, fn)
            mt = mp.get_target(k)

            self._targets[key] = mt
        return self._targets[key]

    def get_mp(self, C: AbsDirPath, fn: AbsFilePath) -> MakefileParsed:
        key = (C, fn)
        if key not in self._files:
            data = read_ustring_from_utf8_file(fn)
            self._files[key] = parse_makefile(C, fn, data)
        return self._files[key]


def get_build_system(C: AbsDirPath, fn: AbsFilePath) -> BuildSystem:
    bs = BuildSystem(_targets={}, _files={})

    go1(C, fn, bs=bs, only=None)

    return bs


def go1(
    C: DirPath, fn: AbsFilePath, bs: BuildSystem, only: Optional[List[TargetName]],
):
    mp = bs.get_mp(C, fn)

    G = get_digraph(mp)

    if only:
        selected = set()
        for _ in only:
            if not _ in G:
                msg = "Cannot find node in G."
                raise ZValueError(msg, C=C, fn=fn, node=_, available=list(G.nodes))
            selected.add(_)
            selected.update(descendants(G, _))
    else:
        selected = set(G.nodes)

    # tt = {}
    for k, v in mp.targets.items():
        if k not in selected:
            continue

        bs.get_make_target(C, fn, k)


def chill(depends_on: List[str]):
    return get_md5("-".join(map(repr, depends_on)))


def make_bridge_main(args=None):
    parser = argparse.ArgumentParser(prog="zuper-make",)
    parser.add_argument("-o", "--out", default="out-zuper-make")
    parser.add_argument("-c", "--command", default=None)
    parser.add_argument("--gui", default=False, action="store_true")
    parser.add_argument("--retries", default=1, type=int, help="Number of times to retry")
    parser.add_argument(
        "--draw-deps", default=False, action="store_true", help="Creates the depedency graph using dot."
    )

    parsed, extra = parser.parse_known_args(args=args)
    fn = cast(AbsFilePath, os.path.abspath(extra[0]))
    retries = parsed.retries
    C: DirPath
    C = cast(DirPath, os.path.dirname(fn))
    if not C:
        C = cast(DirPath, os.path.abspath(os.getcwd()))

    bs = get_build_system(C, fn)
    # logger.info(bs=bs)

    from compmake import Context

    db = parsed.out
    context = Context(db=db)

    jobs = {}

    def get_job_promise(
        C_: AbsDirPath,
        fn_: AbsFilePath,
        target_: TargetName,
        extra_dependencies: List[Tuple[AbsDirPath, AbsFilePath, TargetName]],
        stack: Tuple[Tuple[AbsDirPath, AbsFilePath, TargetName], ...],
    ) -> Promise:

        key = (C_, fn_, target_)
        stack2 = stack + (key,)
        if key in extra_dependencies:
            raise ZValueError("problem", key=key, extra_dependencies=extra_dependencies, stack=stack2)
        # logger.info(key=key, extra_dependencies=len(extra_dependencies))
        if key not in jobs:
            m: MakeTarget = bs.get_make_target(C_, fn_, target_)
            # logger.info(m=m, extra_dependencies=extra_dependencies)

            depends_on = []

            for d, e, f in extra_dependencies:
                jp = get_job_promise(d, e, f, extra_dependencies=[], stack=stack2)
                depends_on.append(jp)

            for d, e, f in m.dependencies:
                # logger.debug(key=key, dep=(d,e,f))
                if (d, e) == (C_, fn_):
                    jp = get_job_promise(d, e, f, extra_dependencies=extra_dependencies, stack=stack2)
                else:
                    extra_deps = extra_dependencies + m.get_local_deps2()
                    jp = get_job_promise(d, e, f, extra_dependencies=extra_deps, stack=stack2)
                depends_on.append(jp)

            if fn_ == fn:
                job_id = target_
            else:
                job_id = os.path.basename(C_) + "--" + target_

            # if not m.other_commands:
            #     job_id += '-c'
            ignore_others: List[TargetName] = m.get_local_deps()
            if m.other_commands:
                jobs[key] = context.comp(
                    run_one_command,
                    C=m.cwdir,
                    fnrel=m.filename,
                    target=m.target,
                    ignore_others=ignore_others,
                    depends_on=depends_on,
                    retries=retries,
                    job_id=job_id,
                )
            else:
                jobs[key] = context.comp(chill, depends_on=depends_on, job_id=job_id,)

        res = jobs[key]
        # check_isinstance(res, str)
        return res

    mp = bs.get_mp(C, fn)
    for k in mp.targets:
        get_job_promise(C, fn, k, extra_dependencies=[], stack=())

    # if parsed.draw_deps:
    #     write_dot(G, "deps.dot")
    #     subprocess.check_call(["dot", "-Tpdf", "-odeps.pdf", "deps.dot"])

    if parsed.command:

        context.batch_command(parsed.command)
    else:
        if parsed.gui:
            compmake_console_gui(context)
        else:
            context.compmake_console()


def run_one_command(
    C: str,
    fnrel: str,
    target: TargetName,
    ignore_others: List[TargetName],
    depends_on: List[str],
    retries: int,
) -> str:
    _ = depends_on
    # logger.info(cwd=C, fnrel=fnrel, target=target, ignore_others=ignore_others)

    command = ["make", "-f", fnrel, target]
    for i in ignore_others:
        command.extend(["-o", i])
    for retry in range(retries):

        commands = " ".join(command)
        s = f"""
Running:                [try {retry + 1} of {retries}]

    cd {C} && \\
    {commands}

    """
        logger.info(s)

        all_output = []
        p = Popen(command, cwd=C, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True)
        for line in p.stdout:
            sys.stdout.write(line)
            all_output.append(line)
            sys.stdout.flush()
        p.wait(timeout=60)
        retcode = p.returncode
        # logger.info(f'retcode: {retcode}')
        if retcode == 0:
            break

        all_output_s = "".join(all_output)
        temp_error = "i/o" in all_output_s

        if temp_error and (retry != retries - 1):
            msg = f"Command failed with {retcode}. Will retry."
            logger.warning(msg, e=traceback.format_exc())
        else:
            msg = f"Command failed with {retcode} ({retries} times)."
            raise ZException(msg, all_output=all_output_s)
    return datetime.now().isoformat()


def remove_empty_and_comments(lines: List[str]):
    for line in lines:
        line = remove_comments(line)
        line_stripped = line.strip(" ")
        if line_stripped:
            yield line_stripped


def get_digraph(m: MakefileParsed) -> DiGraph:
    G = DiGraph()
    for target in m.targets:
        G.add_node(target)
    for target, info in m.targets.items():

        for _ in info.dependencies:
            G.add_edge(target, _)
    return G


def parse_makefile(cwdir: AbsDirPath, fn: AbsFilePath, data: str) -> MakefileParsed:
    var_values = {}

    data = data.replace("\\\n", " ")
    lines = data.split("\n")
    lines = list(remove_empty_and_comments(lines))

    assignments: Dict[str, str] = {}
    conditional_assignments = {}
    targets: Dict[TargetName, TargetInfo] = {}
    templates: Dict[TargetName, TargetInfo] = {}

    def next_line() -> str:
        return lines[0]

    def pop() -> str:
        return lines.pop(0)

    def replace_variables(s: str) -> str:
        for k, v in var_values.items():
            s = s.replace(f"$({k})", v)
        return s

    while lines:
        line = pop()

        line = replace_variables(line)

        if line.startswith("\t"):
            continue
        if "?=" in line:
            key, _, value = line.partition("?=")
            key = key.strip()
            value = value.strip()
            conditional_assignments[key] = value
            var_values[key] = os.environ.get(key, value)
            continue

        elif "=" in line:
            key, _, value = line.partition("=")
            key = key.strip()
            value = value.strip()
            assignments[key] = value
            continue
        elif "include" in line:
            # line = line.replace("include", "").strip()

            continue
        elif ":" in line:
            target, _, deps = line.partition(":")

            target = cast(TargetName, target.strip())
            deps = cast(List[TargetName], deps.strip().split())
            commands = []
            while lines and next_line().startswith("\t"):
                line = pop()
                line = line.strip()
                line = replace_variables(line)
                if not line:
                    continue
                c = interpret_command(cwdir, line)
                # logger.info(line=line, c=c)
                commands.append(c)
            if "%" in target:
                templates[target] = TargetInfo(dependencies=deps, commands=commands)
            else:
                targets[target] = TargetInfo(dependencies=deps, commands=commands)
        else:
            raise ZValueError(line=line)

    return MakefileParsed(
        cwdir,
        fn,
        assignments=assignments,
        conditional_assignments=conditional_assignments,
        targets=targets,
        templates=templates,
    )


def interpret_command(cwdir: AbsDirPath, line: str):
    prefix = "$(MAKE) -C"
    if line.startswith(prefix):
        rest = line.replace(prefix, "")
        others = rest.strip().split()
        c = os.path.join(cwdir, others[0])
        # TODO: allow different files (-f option)
        filename = cast(AbsFilePath, os.path.join(c, "Makefile"))

        the_dir = cast(AbsDirPath, os.path.join(cwdir, others[0]))
        return MakeC(the_dir, filename, cast(List[TargetName], others[1:]))
    else:
        return RegularCommand(line)


def remove_comments(line: str) -> str:
    try:
        i = line.index("#")
    except ValueError:
        return line
    else:
        return line[:i]


if __name__ == "__main__":
    make_bridge_main()



from ..structures import Cache
from ..jobs import get_job_cache
from nose.tools import istest
from .compmake_test import CompmakeTest


def job_success(*args, **kwargs):
    pass


def job_failure(*args, **kwargs):
    raise ValueError("This job fails")


def check_job_states(db, **expected):
    for job_id, expected_status in expected.items():
        status = get_job_cache(job_id, db=db).state
        if status != expected_status:
            msg = "For job %r I expected status %s but got status %s." % (job_id, expected_status, status)
            raise Exception(msg)


@istest
class TestBlocked(CompmakeTest):
    def mySetUp(self):
        pass

    def testAdding(self):
        comp = self.comp

        A = comp(job_success, job_id="A")
        B = comp(job_failure, A, job_id="B")
        comp(job_success, B, job_id="C")

        def run():
            self.cc.batch_command("make")

        self.assertMakeFailed(run, nfailed=1, nblocked=1)

        check_job_states(self.db, A=Cache.DONE, B=Cache.FAILED, C=Cache.BLOCKED)



"""
    A Job represents the computation as passed by the user.
    It contains only the "action" but not the state.
    (The state of the computation is represented by a Cache object.)

    A Cache object can be in one of the following states:

    *) non-existent / or NOT_STARTED
       (no difference between these states)

    *) IN_PROGRESS: The yielding mechanism is taking care of
       the incremental computation.

       computation:  current computation
       user_object:  None / invalid
       timestamp:    None / timestamp
       tmp_result:   set to the temporary result (if any)

       In this state, we also publish a progress report.

    *) DONE:  The computation has been completed

       computation:  current computation
       user_object: the result of the computation
       timestamp:   when computation was completed
       timetaken:   time taken by the computation
       tmp_result:  None


    *) FAILED
       The computation has failed for some reason

       computation:  failed computation

    Note that user_object and tmp_result are stored separately
    from the Cache element.

    DB Layout:

        'job_id:computation'       Job object
        'job_id:cache'             Cache object
        'job_id:user_object'       Result of the computation
        'job_id:user_object_tmp'



    Up-to-date or not?
    =================

    Here we have to be careful because of the fact that we have
    the special state MORE_REQUESTED.
    Is it a computation done if MORE_REQUESTED? Well, we could say
    no, because when more is completed, the parents will need to be
    redone. However, the use case is that:
    1) you do the all computation
    2) you explicity ask MORE for some targets
    3) you explicitly ask to redo the parents of those targets
    Therefore, a MORE_REQUESTED state is considered as uptodate.


"""
from dataclasses import dataclass
from typing import Dict, List, NewType, Optional, Set, Tuple, Union

from zuper_commons.types import describe_value
from zuper_commons.ui import duration_compact

from .utils.pickle_frustration import pickle_main_context_save

__all__ = [
    "Promise",
    "Job",
    "Cache",
    "CMJobID",
    "DBKey",
]

CMJobID = NewType("CMJobID", str)
DBKey = NewType("DBKey", str)


@dataclass
class Promise:
    job_id: CMJobID


class Job:
    job_id: CMJobID
    children: Set
    parents: Set
    needs_conntext: bool
    defined_by: List[CMJobID]
    dynamic_children: dict
    pickle_main_context: object
    command_desc: str

    def __init__(
        self,
        job_id: CMJobID,
        children: Set[CMJobID],
        command_desc: str,
        needs_context: bool = False,
        defined_by: List[CMJobID] = None,
    ):
        """

            needs_context: new facility for dynamic jobs
            defined_by: name of jobs defining this job dynamically
                        This is the stack of jobs. 'root' is the first.

            children: the direct dependencies
        """
        self.job_id = job_id
        self.children = set(children)
        self.command_desc = command_desc
        self.parents = set()
        self.needs_context = needs_context
        self.defined_by = defined_by
        assert len(defined_by) >= 1, defined_by
        assert defined_by[0] == "root", defined_by
        # str -> set(str), where the key is one
        # of the direct children
        self.dynamic_children = {}

        self.pickle_main_context = pickle_main_context_save()


def same_computation(jobargs1, jobargs2):
    """ Returns boolean, string tuple """
    cmd1, args1, kwargs1 = jobargs1
    cmd2, args2, kwargs2 = jobargs2

    equal_command = cmd1 == cmd2
    equal_args = args1 == args2
    equal_kwargs = kwargs1 == kwargs2

    equal = equal_args and equal_kwargs and equal_command
    if not equal:
        reason = ""

        if not equal_command:
            reason += "* function changed \n"
            reason += f"  - old: {cmd1} \n"
            reason += f"  - new: {cmd2} \n"

            # TODO: can we check the actual code?

        warn = " (or you did not implement proper __eq__)"
        if len(args1) != len(args2):
            reason += f"* different number of arguments ({len(args1)} -> {len(args2)})\n"
        else:
            for i, ob in enumerate(args1):
                if ob != args2[i]:
                    reason += f"* arg #{i} changed {warn} \n"
                    reason += f"  - old: {describe_value(ob)}\n"
                    reason += f"  - old: {describe_value(args2[i])}\n"

        for key, value in kwargs1.items():
            if key not in kwargs2:
                reason += f'* kwarg "{key}" not found\n'
            elif value != kwargs2[key]:
                reason += f'* argument "{key}" changed {warn} \n'
                reason += f"  - old: {describe_value(value)} \n"
                reason += f"  - new: {describe_value(kwargs2[key])} \n"

        # TODO: different lengths

        return False, reason
    else:
        return True, None


class IntervalTimer:
    def __init__(self):
        import time

        self.c0 = time.process_time()
        self.t0 = time.time()
        self.stopped = False

    def stop(self):
        self.stopped = True
        import time

        self.c1 = time.process_time()
        self.t1 = time.time()

    def get_walltime_used(self):
        if not self.stopped:
            raise ValueError("not stopped")
        return self.t1 - self.t0

    def get_cputime_used(self):
        if not self.stopped:
            raise ValueError("not stopped")
        return self.c1 - self.c0

    def walltime_interval(self):
        if not self.stopped:
            raise ValueError("not stopped")
        return self.t0, self.t1

    def __str__(self):
        tms = int((self.t1 - self.t0) * 1000)
        cms = int((self.c1 - self.c0) * 1000)
        return f"Timer(wall {tms} ms cpu {cms} ms)"


StateCode = NewType("StateCode", int)


class Cache:
    # TODO: add blocked

    NOT_STARTED = StateCode(0)
    PROCESSING = StateCode(1)
    FAILED = StateCode(3)
    BLOCKED = StateCode(5)
    DONE = StateCode(4)

    TIMESTAMP_TO_REMAKE = 0.0

    allowed_states: List[StateCode] = [NOT_STARTED, FAILED, DONE, BLOCKED, PROCESSING]

    state2desc: Dict[StateCode, str] = {
        NOT_STARTED: "todo",
        BLOCKED: "blocked",
        FAILED: "failed",
        DONE: "done",
        PROCESSING: "processing",
    }

    stateupdate2color = {
        # (state, uptodate)
        (NOT_STARTED, False): {},
        #     (Cache.NOT_STARTED, False): {'color': 'white', 'attrs': ['concealed']},
        (FAILED, False): {"color": "red"},
        (BLOCKED, True): {"color": "brown"},
        (BLOCKED, False): {"color": "brown"},  # XXX
        (DONE, True): {"color": "green"},
        (DONE, False): {"color": "magenta"},
    }

    state2color = {
        NOT_STARTED: {"color": "yellow"},  # {'attrs': ['dark']},
        #     Cache.IN_PROGRESS: {'color': 'yellow'},
        BLOCKED: {"color": "brown"},
        FAILED: {"color": "red"},
        DONE: {"color": "green"},
    }

    styles = {
        NOT_STARTED: dict(color="yellow"),
        DONE: dict(color="green"),
        PROCESSING: dict(color="blue"),
        FAILED: dict(color="red"),
        BLOCKED: dict(color="brown"),
        "ready": dict(color="yellow"),
    }

    glyphs = {
        NOT_STARTED: "?",
        DONE: "✔",
        PROCESSING: "⚙",
        FAILED: "✗",
        BLOCKED: "⯃",
        "ready": "⛦",
        "todo": "⌖",
    }

    state: StateCode
    timestamp_started: Optional[float]
    """ time start """
    timestamp: float
    """ time end """

    int_load_results: Optional[IntervalTimer]
    int_make: Optional[IntervalTimer]
    int_compute: Optional[IntervalTimer]
    int_save_results: Optional[IntervalTimer]
    int_gc: Optional[IntervalTimer]
    jobs_defined: Set[CMJobID]
    hashes_dependencies: Dict[str, object]
    exception: Optional[str]
    backtrace: Optional[str]
    captured_stdout: Optional[str]
    captured_stderr: Optional[str]
    walltime_used: Optional[float]
    cputime_used: Optional[float]

    def __init__(self, state: StateCode):
        assert state in Cache.allowed_states
        self.state = state
        # time start
        self.timestamp_started = None
        # time end
        self.timestamp = 0.0

        # Hash for dependencies when this was computed
        self.hashes_dependencies = {}

        self.jobs_defined = set()

        # in case of failure
        self.exception = None  # a short string
        self.backtrace = None  # a long string
        self.captured_stdout = None
        self.captured_stderr = None

        # total
        self.cputime_used = None
        self.walltime_used = None

        # phases
        # make = load + compute + save

        self.int_make = None
        self.int_load_results = None
        self.int_compute = None
        self.int_save_results = None
        self.int_gc = None

    def __repr__(self):
        return "Cache(%s;%s;cpu:%s;wall:%s)" % (
            Cache.state2desc[self.state],
            self.timestamp,
            self.cputime_used,
            self.walltime_used,
        )


def cache_has_large_overhead(cache: Cache) -> bool:
    overhead = (
        cache.int_load_results.get_walltime_used()
        + cache.int_save_results.get_walltime_used()
        + cache.int_gc.get_walltime_used()
    )
    return overhead - cache.int_make.get_walltime_used() > 1.0


def timing_summary(cache: Cache) -> str:
    dc = duration_compact
    s = "%7s (L %s C %s GC %s S %s)" % (
        dc(cache.int_make.get_walltime_used()),
        dc(cache.int_load_results.get_walltime_used()),
        dc(cache.int_compute.get_walltime_used()),
        dc(cache.int_gc.get_walltime_used()),
        dc(cache.int_save_results.get_walltime_used()),
    )
    return s


class ProgressStage:
    def __init__(self, name: str, iterations: Tuple[Union[float, int], Union[float, int]], iteration_desc):
        self.name = name
        self.iterations = iterations
        self.iteration_desc = iteration_desc
        # We keep track of when to send the event
        self.last_broadcast = None

    def __str__(self):
        return f"[{self.name} {self.iterations} {self.iteration_desc}]"

    def was_finished(self):
        # allow off-by-one conventions
        if isinstance(self.iterations[1], int):
            return self.iterations[0] >= self.iterations[1] - 1
        else:
            return self.iterations[0] >= self.iterations[1]



""" The actual interface of some commands in commands.py """
from typing import List, Tuple

from compmake.exceptions import CompmakeBug
from compmake.jobs.storage import all_jobs, get_job, job_exists

from ..jobs import children, direct_children, direct_parents, parents, parse_job_list
from ..ui import COMMANDS_ADVANCED, ui_command
from ..ui.visualization import ui_error


@ui_command(section=COMMANDS_ADVANCED, alias="check-consistency")
def check_consistency(args, context, cq, raise_if_error=False):  # @ReservedAssignment
    """ Checks in the DB that the relations between jobs are consistent. """

    db = context.get_compmake_db()

    # Do not use cq
    if not args:
        job_list = all_jobs(db=db)
    else:
        job_list = parse_job_list(args, context=context)

    job_list = list(job_list)
    # print('Checking consistency of %d jobs.' % len(job_list))
    errors = {}
    for job_id in job_list:
        try:
            ok, reasons = check_job(job_id, context)
            if not ok:
                errors[job_id] = reasons
        except CompmakeBug as e:
            errors[job_id] = ["bug: %s" % e]

    if errors:
        msg = f"Inconsistency with {len(errors)} jobs:"
        for job_id, es in errors.items():
            msg += "\n- job %r:\n%s" % (job_id, "\n".join(es))

        if raise_if_error:
            raise CompmakeBug(msg)
        else:
            ui_error(context, msg)

    return 0


def check_job(job_id, context) -> Tuple[bool, List[str]]:
    db = context.get_compmake_db()

    job = get_job(job_id, db)
    defined_by = job.defined_by
    assert "root" in defined_by

    dparents = direct_parents(job_id, db=db)
    all_parents = parents(job_id, db=db)
    dchildren = direct_children(job_id, db=db)
    all_children = children(job_id, db=db)

    # print(job_id)
    # print('d children: %s' % dchildren)
    # print('all children: %s' % all_children)

    errors = []

    def e(msg):
        errors.append(msg)

    for defb in defined_by:
        if defb == "root":
            continue
        if not job_exists(defb, db=db):
            s = "%r defined by %r but %r not existing." % (job_id, defined_by, defb)
            e(s)

    for dp in dparents:

        if not job_exists(dp, db=db):
            s = f"Direct parent {dp!r} of {job_id!r} does not exist."
            e(s)
        else:
            if not job_id in direct_children(dp, db=db):
                s = f"{job_id} thinks {dp} is its direct parent;"
                s += f"but {dp} does not think {job_id} is its direct child"
                e(s)

    for ap in all_parents:
        if not job_exists(ap, db=db):
            s = f"Parent {ap!r} of {job_id!r} does not exist."
            e(s)
        else:
            if not job_id in children(ap, db=db):
                e(f"{ap} is parent but no child relation")

    for dc in dchildren:
        if not job_exists(dc, db=db):
            s = f"Direct child {dc!r} of {job_id!r} does not exist."
            e(s)
        else:
            if not job_id in direct_parents(dc, db=db):
                e(f"{dc} is direct child but no direct_parent relation")

    for ac in all_children:
        if not job_exists(ac, db=db):
            s = f"A child {ac!r} of {job_id!r} does not exist."
            e(s)
        else:
            if not job_id in parents(ac, db=db):
                e(f"{ac} is direct child but no parent relation")

    if errors:
        s = f"Inconsistencies for {job_id}:\n"
        s += "\n".join(f"- {msg}" for msg in errors)
        ui_error(context, s)
        return False, errors
    else:
        return True, []



from nose.tools import istest

from .compmake_test import CompmakeTest
from .mockup import mockup_recursive_5


@istest
class TestDynamic2rec(CompmakeTest):
    def test_dynamic1(self):
        mockup_recursive_5(self.cc)
        self.assert_cmd_success("parmake recurse=1;ls")
        self.assertJobsEqual("all", ["r1", "r2", "r3", "r4", "r5"])
        self.assertJobsEqual("done", ["r1", "r2", "r3", "r4", "r5"])



from zuper_commons.types import check_isinstance

__all__ = [
    "result_dict_check",
    "result_dict_raise_if_error",
]


def result_dict_check(res):
    check_isinstance(res, dict)
    # print(res.__repr__().__repr__()) # XXX
    msg = "Invalid result dict"  #% res
    # print('result_dict: %s' % res)
    if "new_jobs" in res:
        assert "new_jobs" in res, msg
        assert "deleted_jobs" in res, msg
        assert "user_object_deps" in res, msg
    elif "fail" in res:
        assert "deleted_jobs" in res, msg
    elif "bug" in res:
        pass
    elif "abort" in res:
        pass
    elif "interrupted" in res:
        assert "deleted_jobs" in res, msg
        pass
    else:
        msg = "Malformed result dict: %s" % res
        raise ValueError(msg)


def result_dict_raise_if_error(res):
    from compmake.exceptions import JobFailed
    from compmake.exceptions import HostFailed
    from compmake.exceptions import CompmakeBug
    from compmake.exceptions import JobInterrupted

    result_dict_check(res)

    if "fail" in res:
        raise JobFailed.from_dict(res)

    if "abort" in res:
        raise HostFailed.from_dict(res)

    if "bug" in res:
        raise CompmakeBug.from_dict(res)

    if "interrupted" in res:
        raise JobInterrupted.from_dict(res)



from .. import logger
from .storage import *
from .progress_imp2 import *
from .queries import *
from .uptodate import *
from .actions import *
from .priority import *
from .manager import *
from .syntax.parsing import *
from .dependencies import *
from .actions_newprocess import *
from .result_dict import *



from copy import deepcopy
from typing import Set

from compmake.exceptions import CompmakeBug
from compmake.jobs import get_job_userobject, job_userobject_exists
from compmake.structures import CMJobID
from zuper_commons.types import raise_wrapped, ZException

__all__ = [
    "substitute_dependencies",
    "collect_dependencies",
    "get_job_userobject_resolved",
]


def get_job_userobject_resolved(job_id: CMJobID, db):
    """ This gets the job's result, and recursively substitute all
    dependencies. """
    ob = get_job_userobject(job_id, db)
    all_deps = collect_dependencies(ob)
    for dep in all_deps:
        if not job_userobject_exists(dep, db):
            msg = f"Cannot resolve {job_id!r}: dependency {dep!r} was not done."
            raise CompmakeBug(msg)
    return substitute_dependencies(ob, db)


def substitute_dependencies(a, db):
    from compmake import Promise

    # XXX: this is a workaround
    if leave_it_alone(a):
        return deepcopy(a)

    if isinstance(a, dict):
        ca = type(a)
        rest = [(k, substitute_dependencies(v, db=db)) for k, v in a.items()]
        try:
            # noinspection PyArgumentList
            res = ca(rest)
            # print('%s->%s' % (a, str(res)))
            return res
        except (BaseException, TypeError) as e:
            msg = "Could not instance something looking like a dict."
            raise ZException(msg, ca=ca) from e

    elif isinstance(a, list):
        # XXX: This fails for subclasses of list
        # noinspection PyArgumentList
        return type(a)([substitute_dependencies(x, db=db) for x in a])
    elif isinstance(a, tuple):
        # First, check that there are dependencies
        deps_in_tuple = collect_dependencies(a)
        if not deps_in_tuple:
            # if not, just return the tuple
            return a
        # XXX: This fails for subclasses of tuples
        assert not isnamedtupleinstance(a), a

        ta = type(a)
        contents = [substitute_dependencies(x, db=db) for x in a]
        try:
            # noinspection PyArgumentList
            return ta(contents)
        except TypeError as e:
            msg = "Cannot reconstruct complex tuple."
            raise_wrapped(ValueError, e, msg, ta=ta, contents=contents)
    elif isinstance(a, Promise):
        s = get_job_userobject(a.job_id, db=db)
        return substitute_dependencies(s, db=db)
    else:
        return a


def collect_dependencies(ob) -> Set[CMJobID]:
    """ Returns a set of dependencies (i.e., Promise objects that
        are mentioned somewhere in the structure """
    from compmake import Promise

    if isinstance(ob, Promise):
        return {ob.job_id}
    else:
        if leave_it_alone(ob):
            return set()

        depends = set()
        if isinstance(ob, (list, tuple)):
            for child in ob:
                depends.update(collect_dependencies(child))
        if isinstance(ob, dict):
            for child in ob.values():
                depends.update(collect_dependencies(child))
        return depends


def leave_it_alone(x):
    """ Returns True for those objects that have trouble with factory methods
        like namedtuples. """
    if isnamedtupleinstance(x):
        return True

    # XXX: this is a workaround
    if type(x).__name__ in ["ObjectSpec"]:
        return True

    return False


def isnamedtupleinstance(x):
    t = type(x)
    b = t.__bases__
    if len(b) != 1 or b[0] != tuple:
        return False
    f = getattr(t, "_fields", None)
    if not isinstance(f, tuple):
        return False
    return all(type(n) == str for n in f)



# -*- coding: utf-8 -*-

import multiprocessing
import os
import random
import signal
import sys
import tempfile
import time
from multiprocessing import Pool, Queue
from typing import Dict

import setproctitle
from future.moves.queue import Empty

from compmake import CompmakeGlobalState
from compmake.events.registrar import broadcast_event, publish
from compmake.exceptions import HostFailed
from compmake.jobs.manager import AsyncResultInterface, Manager
from compmake.jobs.result_dict import result_dict_raise_if_error
from compmake.plugins.backend_pmake.parmake_job2_imp import parmake_job2
from compmake.state import get_compmake_config
from .shared import Shared

__all__ = [
    "MultiprocessingManager",
]

# for some reason it might block on OSX 10.8
ncpus = multiprocessing.cpu_count()


#
# if False:
# # Debug multiprocsssing
#     import logging
#     logger = multiprocessing.log_to_stderr(logging.DEBUG)
#     logger.setLevel(multiprocessing.SUBDEBUG)

#
# def sig_child(signo, frame):
#     # error('Child terminated %s %s' % (signo, frame))
#     pass


class MultiprocessingManager(Manager):
    """ Specialization of Manager for local multiprocessing """

    pool: Pool
    max_num_processing: int
    num_processes: int
    last_accepted: float
    " Time accepted "

    def __init__(self, context, cq, num_processes: int, recurse: bool):
        Manager.__init__(self, context=context, cq=cq, recurse=recurse)
        self.num_processes = num_processes
        self.last_accepted = 0.0

    def process_init(self) -> None:
        Shared.event_queue = Queue(self.num_processes * 1000)
        # info('Starting %d processes' % self.num_processes)
        kwargs = {}

        if sys.hexversion >= 0x02070000:
            # added in 2.7.2
            kwargs["maxtasksperchild"] = 1

        self.pool = Pool(processes=self.num_processes, initializer=worker_initialization, **kwargs)
        self.max_num_processing = self.num_processes

    def get_resources_status(self):
        resource_available = {}

        # Scale up softly
        time_from_last = time.time() - self.last_accepted
        min_interval = get_compmake_config("min_proc_interval")
        if time_from_last < min_interval:
            resource_available["soft"] = (False, "interval: %.2f < %.1f" % (time_from_last, min_interval))
        else:
            resource_available["soft"] = (True, "")

        # only one job at a time
        process_limit_ok = len(self.processing) < self.max_num_processing
        if not process_limit_ok:
            resource_available["nproc"] = (False, "max %d nproc" % (self.max_num_processing))
            # this is enough to continue
            return resource_available
        else:
            resource_available["nproc"] = (True, "")

        # TODO: add disk

        stats = CompmakeGlobalState.system_stats
        if not stats.available():  # psutil not installed
            resource_available["cpu"] = (True, "n/a")
            resource_available["mem"] = (True, "n/a")
        else:
            # avg_cpu = stats.avg_cpu_percent()
            max_cpu = stats.max_cpu_percent()
            cur_mem = stats.cur_phymem_usage_percent()
            cur_swap = stats.cur_virtmem_usage_percent()

            num_processing = len(self.processing)
            if num_processing > 0:  # at least one
                if ncpus > 2:
                    # Do this only for big machines
                    # XXX: assumes we are cpu-bound
                    estimated_cpu_increase = 1.0 / ncpus
                    estimated_cpu = max_cpu + estimated_cpu_increase
                    max_cpu_load = get_compmake_config("max_cpu_load")
                    if max_cpu_load < 100 and estimated_cpu > max_cpu_load:
                        reason = "cpu %d%%, proj %d%% > %d%%" % (max_cpu, estimated_cpu, max_cpu_load)
                        resource_available["cpu"] = (False, reason)
                    else:
                        resource_available["cpu"] = (True, "")

                max_mem_load = get_compmake_config("max_mem_load")
                if cur_mem > max_mem_load:
                    reason = "mem %s > %s" % (cur_mem, max_mem_load)
                    resource_available["mem"] = (False, reason)
                    # print('Memory load too high: %s\n\n' % cpu_load)
                else:
                    resource_available["mem"] = (True, "")

                max_swap = get_compmake_config("max_swap")
                if cur_swap > max_swap:
                    reason = "swap %s > %s" % (cur_swap, max_swap)
                    resource_available["swap"] = (False, reason)
                    # print('Memory load too high: %s\n\n' % cpu_load)
                else:
                    resource_available["swap"] = (True, "")

            # cooperating between parmake instances:
            # to balance the jobs, accept with probability
            # 1 / (1+n), where n is the number of current processes
            if True:
                autobal_after = get_compmake_config("autobal_after")
                n = len(self.processing)
                q = max(0, n - autobal_after)
                probability = 1.0 / (1 + q)
                if random.random() > probability:
                    # Unlucky, let's try next time
                    reason = "after %d, p=%.2f" % (autobal_after, probability)
                    resource_available["autobal"] = (False, reason)
                else:
                    resource_available["autobal"] = (True, "")

        return resource_available

    def can_accept_job(self, reasons_why_not: Dict) -> bool:
        resources = self.get_resources_status()
        some_missing = False
        for k, v in resources.items():
            if not v[0]:
                some_missing = True
                reasons_why_not[k] = v[1]
        if some_missing:
            return False

        self.last_accepted = time.time()
        return True

    def instance_job(self, job_id):
        publish(self.context, "worker-status", job_id=job_id, status="apply_async")
        handle, tmp_filename = tempfile.mkstemp(prefix="compmake", text=True)
        os.close(handle)
        os.remove(tmp_filename)
        async_result = self.pool.apply_async(parmake_job2, [(job_id, self.context, tmp_filename, False)])
        publish(self.context, "worker-status", job_id=job_id, status="apply_async_done")
        return AsyncResultWrap(job_id, async_result, tmp_filename)

    def event_check(self):
        while True:
            try:
                event = Shared.event_queue.get(block=False)  # @UndefinedVariable
                event.kwargs["remote"] = True
                broadcast_event(self.context, event)
            except Empty:
                break

    def process_finished(self):
        # Make sure that all the stuff is read from the queue
        # otherwise some workers will hang
        # http://docs.python.org/library/multiprocessing.html
        print("process_finished")
        self.event_check()
        self.pool.close()
        self.pool.join()
        self.event_check()
        Shared.event_queue.close()

    def cleanup(self):
        if "pool" in self.__dict__:
            try:
                self.pool.terminate()
            except:
                # multiprocessing/pool.py", line 478, in _terminate_pool
                # assert result_handler.is_alive() or len(cache) == 0
                pass


class AsyncResultWrap(AsyncResultInterface):
    """ Wrapper for the async result object obtained by pool.apply_async """

    def __init__(self, job_id, async_result, tmp_filename):
        self.job_id = job_id
        self.async_result = async_result
        self.tmp_filename = tmp_filename

        self.count = 0

    def ready(self):
        self.count += 1
        is_ready = self.async_result.ready()
        #         tmp_filename = self.tmp_filename

        if self.count > 10000 and (self.count % 100 == 0):
            #             if is_ready:
            #                 if not os.path.exists(tmp_filename):
            #                     msg = 'I would have expected tmp_filename
            # to exist.\n %s' % tmp_filename
            #                     error('%s: %s' % (self.job_id, msg))
            #             else:
            #                 if os.path.exists(tmp_filename):
            #                     msg = 'The tmp_filename exists! but job
            # not returned yet.\n %s' % tmp_filename
            #                     error('%s: %s' % (self.job_id, msg))
            #
            #                     if self.count % 100 == 0:
            #                         s = open(tmp_filename).read()
            #                         print('%s: %s: %s ' % (self.job_id,
            # self.count, s))

            if False:
                if self.count % 100 == 0:
                    s = self.read_status()  # @UnusedVariable
                    # print('%70s: %10s  %s         ' % (self.job_id, self.count,
                    #  s))

        # timeout
        if self.count > 100000:
            raise HostFailed(host="localhost", job_id=self.job_id, reason="Timeout", bt="")

        return is_ready

    def read_status(self):
        if os.path.exists(self.tmp_filename):
            with open(self.tmp_filename) as f:
                return f.read()
        else:
            return "(no status)"

    def get(self, timeout=0):  # @UnusedVariable
        res = self.async_result.get(timeout=timeout)
        result_dict_raise_if_error(res)
        return res


def worker_initialization() -> None:
    title = "compmake: worker just created"
    setproctitle.setproctitle(title)

    # http://stackoverflow.com/questions/1408356
    # XXX: temporary looking at interruptions
    signal.signal(signal.SIGINT, signal.SIG_IGN)

    # You can use this to see when a worker start
    # print('Process: ignoring sigint')



import sys
from typing import Callable

from compmake.events import publish, register_handler
from compmake.utils import compmake_colored
from zuper_commons.types import check_isinstance
from zuper_commons.ui import get_colorize_function
from .. import get_compmake_config
from ..utils import get_screen_columns

__all__ = ["compmake_colored", "ui_debug", "ui_error", "ui_info", "ui_message", "ui_warning"]


def clean_console_line(stream):
    s = "\r" + (" " * (get_screen_columns() - 0)) + "\r"  # was : 2
    stream.write(s)


#
# def warning(s):
#     write_message(s, lambda x: compmake_colored(x, "yellow"))


def ui_message(context, s: str):
    publish(context, "ui-message", string=s)


color_it_red = get_colorize_function("#ff0000")
color_it_red("hello")
color_it_green = get_colorize_function("#00ff00")
# print(color_it_green("hello"))

color_it_blue = get_colorize_function("#0000ff")
color_it_pink = get_colorize_function("#ffaaaa")
color_it_yellow = get_colorize_function("#0000ff")


def ui_error(context, s: str):
    publish(context, "ui-message", string=make_colored(s, color_it_red))


def ui_info(context, s: str):
    # write_message(s, lambda x: compmake_colored(x, 'green'))
    publish(context, "ui-message", string=make_colored(s, color_it_green))


def ui_debug(context, s: str):
    publish(context, "ui-message", string=make_colored(s, color_it_pink))


def ui_warning(context, s: str):
    publish(context, "ui-message", string=make_colored(s, color_it_yellow))


class DefaultConsole:
    active = True


def handle_ui_message_console(context, event):
    if not DefaultConsole.active:
        return
    write_message(event.kwargs["string"], lambda x: x)


register_handler("ui-message", handle_ui_message_console)

original_stderr = sys.stderr


def handle_ui_status_summary(context, event):
    if not DefaultConsole.active:
        return
    line = event.kwargs["string"]
    if get_compmake_config("console_status"):
        # if six.PY2:
        #     if isinstance(line, unicode):
        #         line = line.encode('utf-8')
        original_stderr.write(line)

        interactive = get_compmake_config("interactive")
        if interactive:
            original_stderr.write("\r")
        else:
            original_stderr.write("\n")


register_handler("ui-status-summary", handle_ui_status_summary)


def make_colored(s: str, f: Callable[[str], str]) -> str:

    lines = s.rstrip().split("\n")
    lines2 = []
    for l in lines:

        lines2.append(f(l))
    res = "\n".join(lines2)
    return res


#
# def error(s: str):
#     write_message(s, lambda x: compmake_colored(x, "red"))
#
#
# def user_error(s: str):  # XXX: what's the difference with above?
#     write_message(s, lambda x: compmake_colored(x, "red"))
#
#
# def info(s: str):
#     # write_message(s, lambda x: compmake_colored(x, 'green'))
#     write_message(s, lambda x: compmake_colored(x, "green"))
#
#
# def debug(s: str):  # XXX: never used?
#     write_message(s, lambda x: compmake_colored(x, "magenta"))


def write_message(s: str, formatting: Callable[[str], str]):
    check_isinstance(s, str)
    from ..utils import pad_to_screen

    from .. import CompmakeGlobalState

    stdout = CompmakeGlobalState.original_stdout

    lines = s.rstrip().split("\n")

    from compmake.plugins.console_output import write_line_endl_w

    for l in lines:
        l = formatting(l)
        s = pad_to_screen(l)
        write_line_endl_w(s, stdout)



from typing import Set

from compmake.structures import CMJobID

from ..events import register_handler

__all__ = [
    "Tracker",
]


class Tracker:
    """ This class keeps track of the status of the computation.
        It listens to progress events. """

    processing: Set[CMJobID]
    targets: Set[CMJobID]
    all_targets: Set[CMJobID]
    todo: Set[CMJobID]
    failed: Set[CMJobID]
    ready: Set[CMJobID]
    blocked: Set[CMJobID]
    done: Set[CMJobID]
    wait_reasons: dict

    def __init__(self):
        register_handler("job-progress", self.event_job_progress)
        register_handler("job-progress-plus", self.event_job_progress_plus)
        register_handler("manager-progress", self.event_manager_progress)
        register_handler("manager-loop", self.event_manager_loop)
        register_handler("manager-wait", self.event_manager_wait)

        self.processing = set()
        self.targets = set()
        self.all_targets = set()
        self.todo = set()
        self.failed = set()
        self.ready = set()
        self.blocked = set()
        self.done = set()
        # Status of jobs in "processing" state
        self.status = {}
        self.status_plus = {}
        self.nloops = 0
        self.wait_reasons = {}

    def event_manager_wait(self, context, event):
        self.wait_reasons = event.reasons

    def event_manager_loop(self, context, event):
        self.nloops += 1

    def event_job_progress(self, context, event):
        """ Receive news from the job """
        # attrs = ['job_id', 'host', 'done', 'progress', 'goal']
        stat = f"{event.progress}/{event.goal}"
        self.status[event.job_id] = stat

    def event_job_progress_plus(self, context, event):
        self.status_plus[event.job_id] = event.stack
        if len(event.stack) > 0:
            i, n = event.stack[0].iterations
            if isinstance(n, int):
                stat = f"{i + 1}/{n}"
            else:
                perc = i * 100.0 / n
                stat = f"{perc:.1f}%"
        else:
            stat = "-"
        self.status[event.job_id] = stat

    def event_manager_progress(self, context, event):
        """ Receive progress message (updates processing) """
        # attrs=['targets', 'done', 'todo', 'failed', 'ready', 'processing']
        self.processing = event.processing
        self.targets = event.targets
        self.todo = event.todo
        self.all_targets = event.all_targets
        self.failed = event.failed
        self.ready = event.ready
        self.done = event.done
        self.blocked = event.blocked

        # Put unknown for new jobs
        for job_id in self.processing:
            if not job_id in self.status:
                self.status[job_id] = "-"
            if not job_id in self.status:
                self.status_plus[job_id] = "-"

        # Remove completed jobs from status
        for job_id in list(self.status.keys()):
            if not job_id in self.processing:
                del self.status[job_id]

        for job_id in list(self.status_plus.keys()):
            if not job_id in self.processing:
                del self.status_plus[job_id]


# print('Processing is now %r and status is %r %r' %
#              (self.processing, self.status.keys(), self.status_plus.keys()))



""" The actual interface of some commands in commands.py """
import os
from time import time
from typing import Dict, List, Sequence, Tuple

from compmake.constants import CompmakeConstants
from compmake.jobs import parse_job_list
from compmake.jobs.storage import (
    job_args_sizeof,
    job_cache_exists,
    job_cache_sizeof,
    job_userobject_exists,
    job_userobject_sizeof,
)
from compmake.jobs.syntax.parsing import is_root_job
from compmake.structures import Cache, cache_has_large_overhead, timing_summary
from compmake.ui import compmake_colored, ui_command, VISUALIZATION
from compmake.ui.visualization import ui_message
from compmake.utils.table_formatter import TableFormatter
from compmake.utils.terminal_size import get_screen_columns
from zuper_commons.ui import duration_compact

# red, green, yellow, blue, magenta, cyan, white.


# if False:
#     format_utility_job = dict(color="white", attrs=["concealed"])
#     format_separator = dict(color="white", attrs=["concealed"])
#     format_when = dict(color="white", attrs=["concealed"])
# else:
format_utility_job = dict()
format_separator = dict()
format_when = dict()


@ui_command(section=VISUALIZATION, alias="list")
def ls(args, context, cq, complete_names=False, reason=False, all_details=False):  # @ReservedAssignment
    """
        Lists the status of the given jobs (or all jobs if none specified
        specified).

        Options:

            ls complete_names=1   # do not abbreviate names
            ls reason=1  # show why jobs are not uptodate
    """

    if not args:
        job_list = cq.all_jobs()
    else:
        job_list = parse_job_list(tokens=args, context=context, cq=cq)

    job_list = list(job_list)
    CompmakeConstants.aliases["last"] = job_list
    list_jobs(context, job_list, cq=cq, complete_names=complete_names, reason=reason, all_details=all_details)
    return 0


# @contract(objects="seq[N](unicode)", returns="tuple(unicode, list[N](unicode), unicode)")
def minimal_names(objects: Sequence[str]) -> Tuple[str, List[str], str]:
    """
        Converts a list of object IDs to a minimal non-ambiguous list of names.

        For example, the names: ::

            test_learn_fast_10
            test_learn_slow_10
            test_learn_faster_10

        is converted to: ::

            fast
            slow
            faster

        Returns prefix, minimal, postfix
    """
    if len(objects) == 1:
        return "", list(objects), ""

    # find the common prefix
    prefix = os.path.commonprefix(objects)
    # invert strings
    objects_i = [o[::-1] for o in objects]
    # find postfix
    postfix = os.path.commonprefix(objects_i)[::-1]
    #     print(objects)
    #     print('prefix: %r post: %r' % (prefix, postfix))
    n1 = len(prefix)
    n2 = len(postfix)
    # remove it
    minimal = [o[n1 : len(o) - n2] for o in objects]

    # recreate them to check everything is ok
    objects2 = [prefix + m + postfix for m in minimal]

    # print objects, objects2
    assert objects == objects2, (prefix, minimal, postfix)
    return prefix, minimal, postfix


def list_jobs(context, job_list, cq, complete_names=False, all_details=False, reason=False):
    job_list = list(job_list)
    # print('%s jobs in total' % len(job_list))
    if not job_list:
        string = "No jobs found."
        ui_message(context, string)
        return

    # maximum job length

    max_len = 100

    def format_job_id(ajob_id):
        if complete_names or len(ajob_id) < max_len:
            return ajob_id
        else:
            b = 15
            r = max_len - b - len(" ... ")
            return ajob_id[:15] + " ... " + ajob_id[-r:]

    # abbreviates the names
    #     if not complete_names:
    #         prefix, abbreviated, postfix = minimal_names(job_list)
    #         job_list = abbreviated

    jlen = max(len(format_job_id(x)) for x in job_list)

    cpu_total = []
    wall_total = []

    tf = TableFormatter(sep="  ")

    for job_id in job_list:
        tf.row()

        cache = cq.get_job_cache(job_id)

        # TODO: only ask up_to_date if necessary
        up, up_reason, up_ts = cq.up_to_date(job_id)

        job = cq.get_job(job_id)

        is_root = is_root_job(job)
        if not is_root:
            msg = (job_id, job, job.defined_by)
            assert len(job.defined_by) >= 1, msg
            assert job.defined_by[0] == "root", msg

            level = len(job.defined_by) - 1
            assert level >= 1
            tf.cell("%d" % level)
        else:
            tf.cell("")

        if job.needs_context:
            tf.cell("d")
        else:
            tf.cell("")

        job_name_formatted = format_job_id(job_id).ljust(jlen)

        # de-emphasize utility jobs
        is_utility = "context" in job_id or "dynrep" in job_id
        if is_utility:
            job_name_formatted = compmake_colored(job_name_formatted, **format_utility_job)

        tf.cell(format_job_id(job_id))

        tag = Cache.state2desc[cache.state]

        k = (cache.state, up)
        assert k in Cache.stateupdate2color, "I found strange state %s" % str(k)

        tag_s = compmake_colored(tag, **Cache.stateupdate2color[k])
        if not up and cache.state in [Cache.DONE, Cache.FAILED]:
            tag_s += "*"
        tf.cell(tag_s)

        if reason:
            tf.cell(up_reason)
            tf.cell(duration_compact(time() - up_ts))

        db = context.get_compmake_db()
        sizes = get_sizes(job_id, db=db)
        size_s = format_size(sizes["total"])
        tf.cell(size_s)

        if cache.state in [Cache.DONE]:
            wall_total.append(cache.walltime_used)
            cpu = cache.cputime_used
            cpu_total.append(cpu)

            if cpu > 5 or cache_has_large_overhead(cache) or all_details:  # TODO: add param
                # s_cpu = duration_compact(cpu)
                s_cpu = timing_summary(cache)
            else:
                s_cpu = ""

            tf.cell(s_cpu)

        else:
            tf.cell("")  # cpu

        if cache.state in [Cache.DONE, Cache.FAILED]:
            when = duration_compact(time() - cache.timestamp)
            when_s = f"({when} ago)"

            when_s = compmake_colored(when_s, **format_when)

            tf.cell(when_s)
        else:
            tf.cell("")  # when

    tf.done()

    ind = "  "

    do_one_column = len(job_list) <= 5

    if do_one_column:
        for line in tf.get_lines():
            string = ind + line
            ui_message(context, string)
    else:
        linewidth = get_screen_columns()
        # print('*'*linewidth)
        # sep = '   ' + compmake_colored('|', color='white', attrs=['dark']) + '   '
        sep = "   " + compmake_colored("|", **format_separator) + "   "

        for line in tf.get_lines_multi(linewidth - len(ind), sep=sep):
            string = ind + line
            ui_message(context, string)
    if cpu_total:
        cpu_time = duration_compact(sum(cpu_total))
        wall_time = duration_compact(sum(wall_total))
        string = f" total {len(job_list)} jobs   CPU time: {cpu_time}   wall: {wall_time}"
        # print(scpu)
        ui_message(context, string)


def format_size(nbytes: int) -> str:
    if nbytes == 0:
        return ""
    if nbytes < 1024 * 1024:  # TODO: add param
        return ""
    mb = float(nbytes) / (1024 * 1024)
    return f"{mb:.2f} MB"


def get_sizes(job_id, db) -> Dict:
    """ Returns byte sizes for jobs pieces.

        Returns dict with keys 'args','cache','result','total'.
    """
    res = {}
    res["args"] = job_args_sizeof(job_id, db)

    if job_cache_exists(job_id, db):
        res["cache"] = job_cache_sizeof(job_id, db)
    else:
        res["cache"] = 0

    if job_userobject_exists(job_id, db):
        res["result"] = job_userobject_sizeof(job_id, db)
    else:
        res["result"] = 0

    res["total"] = res["cache"] + res["args"] + res["result"]
    return res



import os

from compmake.constants import CompmakeConstants
from compmake.exceptions import CompmakeBug, JobFailed
from compmake.utils import safe_pickle_load, which
from system_cmd import system_cmd_result
from zuper_commons.text import indent
from .result_dict import result_dict_check

__all__ = [
    "result_dict_check",
    "parmake_job2_new_process",
]


def parmake_job2_new_process(args):
    """ Starts the job in a new compmake process. """
    (job_id, context) = args
    compmake_bin = which("compmake")

    db = context.get_compmake_db()
    storage = db.basepath  # XXX:
    where = os.path.join(storage, "parmake_job2_new_process")
    if not os.path.exists(storage):
        try:
            os.makedirs(storage)
        except:
            pass

    out_result = os.path.join(where, "%s.results.pickle" % job_id)
    out_result = os.path.abspath(out_result)
    cmd = [compmake_bin, storage]

    # from contracts import all_disabled, indent
    # if not all_disabled():
    #     cmd += ["--contracts"]

    cmd += [
        "--status_line_enabled",
        "0",
        "--colorize",
        "0",
        "-c",
        "make_single out_result=%s %s" % (out_result, job_id),
    ]

    cwd = os.getcwd()
    cmd_res = system_cmd_result(
        cwd,
        cmd,
        display_stdout=False,
        display_stderr=False,
        raise_on_error=False,
        capture_keyboard_interrupt=False,
    )
    ret = cmd_res.ret

    if ret == CompmakeConstants.RET_CODE_JOB_FAILED:  # XXX:
        msg = "Job %r failed in external process" % job_id
        msg += indent(cmd_res.stdout, "stdout| ")
        msg += indent(cmd_res.stderr, "stderr| ")

        res = safe_pickle_load(out_result)
        os.unlink(out_result)
        result_dict_check(res)

        raise JobFailed.from_dict(res)

    elif ret != 0:
        msg = "Host failed while doing %r" % job_id
        msg += "\n cmd: %s" % " ".join(cmd)
        msg += "\n" + indent(cmd_res.stdout, "stdout| ")
        msg += "\n" + indent(cmd_res.stderr, "stderr| ")
        raise CompmakeBug(msg)  # XXX:

    res = safe_pickle_load(out_result)
    os.unlink(out_result)
    result_dict_check(res)
    return res



# -*- coding: utf-8 -*-

import time

from compmake.exceptions import CompmakeException, JobFailed
from compmake.jobs.dependencies import collect_dependencies
from compmake.jobs.job_execution import get_cmd_args_kwargs
from compmake.jobs.result_dict import result_dict_check
from compmake.jobs.storage import get_job, get_job_cache, set_job_cache, set_job_userobject
from compmake.state import get_compmake_config
from compmake.structures import Cache
from zuper_commons.types import check_isinstance
from .logging_imp import disable_logging_if_config

__all__ = [
    "mvac_job",
]


def mvac_instance(db, job_id, volumes, cwd):
    import multyvac

    layer = get_compmake_config("multyvac_layer")
    if not layer:
        layer = None

    command, args, kwargs = get_cmd_args_kwargs(job_id=job_id, db=db)

    core = get_compmake_config("multyvac_core")
    multyvac_job_id = multyvac.submit(
        command, *args, _layer=layer, _vol=volumes, _name=job_id, _core=core, **kwargs
    )
    multyvac_job = multyvac.get(multyvac_job_id)
    return multyvac_job


# @contract(args="tuple(str, *,  str, bool, list, str)")
def mvac_job(args):
    """
    args = tuple job_id, context,  queue_name, show_events

    Returns a dictionary with fields "user_object", "new_jobs", 'delete_jobs'.
    "user_object" is set to None because we do not want to
    load in our thread if not necessary. Sometimes it is necessary
    because it might contain a Promise.

    """
    job_id, context, event_queue_name, show_output, volumes, cwd = args  # @UnusedVariable
    check_isinstance(job_id, str)
    check_isinstance(event_queue_name, str)

    # Disable multyvac logging
    disable_logging_if_config(context)

    db = context.get_compmake_db()
    job = get_job(job_id=job_id, db=db)

    if job.needs_context:
        msg = "Cannot use multyvac for dynamic job."
        raise CompmakeException(msg)

    time_start = time.time()

    multyvac_job = mvac_instance(db, job_id, volumes, cwd)
    multyvac_job.wait()

    errors = [multyvac_job.status_error, multyvac_job.status_killed]
    if multyvac_job.status in errors:
        e = "Multyvac error (status: %r)" % multyvac_job.status
        bt = str(multyvac_job.stderr)

        cache = Cache(Cache.FAILED)
        cache.exception = e
        cache.backtrace = bt
        cache.timestamp = time.time()
        cache.captured_stderr = str(multyvac_job.stderr)
        cache.captured_stdout = str(multyvac_job.stdout)
        set_job_cache(job_id, cache, db=db)

        raise JobFailed(job_id=job_id, reason=str(e), bt=bt)

    user_object = multyvac_job.result

    user_object_deps = collect_dependencies(user_object)
    set_job_userobject(job_id, user_object, db=db)

    cache = get_job_cache(job_id, db=db)
    cache.captured_stderr = str(multyvac_job.stderr)
    cache.captured_stdout = str(multyvac_job.stdout)

    cache.state = Cache.DONE
    cache.timestamp = time.time()
    walltime = cache.timestamp - time_start
    cache.walltime_used = walltime
    cache.cputime_used = multyvac_job.cputime_system
    cache.host = "multyvac"
    cache.jobs_defined = set()
    set_job_cache(job_id, cache, db=db)

    result_dict = dict(
        user_object=user_object, user_object_deps=user_object_deps, new_jobs=[], deleted_jobs=[]
    )
    result_dict_check(result_dict)
    return result_dict



def f(x):
    return x * 2


def statistics(res):
    return sum(res)


def schedule(context, params):
    jobs = [context.comp(f, x=p) for p in params]
    summary = context.comp(statistics, jobs)
    # returns a job "promise", not a value!
    return summary


def report(summary):
    print("The sum is: %r" % summary)


def mockup_dyn4(context):
    summary = context.comp_dynamic(schedule, [42, 43, 44])
    context.comp(report, summary)



import unittest
from tempfile import mkdtemp

from nose.tools import istest

from compmake.context import Context
from compmake.jobs import CMJobID, get_job
from compmake.storage.filesystem import StorageFilesystem


def g():
    return 2


def f(context):
    return context.comp(g)


def e(context):
    return context.comp_dynamic(f)


def h(i):
    assert i == 2


def define_jobs(root):
    db = StorageFilesystem(root, compress=True)
    cc = Context(db=db)
    cc.comp(h, cc.comp_dynamic(e))
    return db, cc


@istest
class TestDelegation5(unittest.TestCase):
    """
        Here's the problem: when the master are overwritten then
        the additional dependencies are lost.
    """

    def test_delegation_5(self):
        root = mkdtemp()
        db, cc = define_jobs(root)
        job0 = get_job(CMJobID("h"), db)
        self.assertEqual(job0.children, {"e"})

        cc.batch_command("make;ls")

        job = get_job(CMJobID("h"), db)
        self.assertEqual(job.children, {"e", "e-f", "e-f-g"})
        print("parents: %s" % job.parents)
        print("children: %s" % job.children)

        cc.batch_command("ls")
        cc.batch_command("check_consistency raise_if_error=1")
        # Now just define h again
        db, cc = define_jobs(root)
        cc.batch_command("check_consistency raise_if_error=1")
        job2 = get_job(CMJobID("h"), db)
        self.assertEqual(job2.children, {"e", "e-f", "e-f-g"})



import os

__all__ = [
    "which",
]


def is_exe(fpath):
    return os.path.exists(fpath) and os.access(fpath, os.X_OK)


def ext_candidates(fpath):
    yield fpath
    for ext in os.environ.get("PATHEXT", "").split(os.pathsep):
        yield fpath + ext


def which(program):
    """ Returns string or raise ValueError. """
    PATH = os.environ["PATH"]
    PATHs = PATH.split(os.pathsep)

    fpath, _ = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in PATHs:
            exe_file = os.path.join(path, program)
            for candidate in ext_candidates(exe_file):
                if is_exe(candidate):
                    return candidate

    msg = "Could not find program %r." % program
    msg += "\n paths = %s" % PATH
    raise ValueError(msg)



import time

from compmake.constants import DefaultsToConfig
from compmake.jobs import Cache, IntervalTimer, mark_to_remake, set_job_cache, set_job_userobject, top_targets
from compmake.ui import ACTIONS, ask_if_sure_remake, raise_error_if_manager_failed, ui_command
from .manager_local import ManagerLocal

__all__ = [
    "make",
    "rmake",
]


@ui_command(section=ACTIONS, dbchange=True)
def make(
    job_list,
    context,
    cq,
    echo: bool = DefaultsToConfig("echo"),
    new_process: bool = DefaultsToConfig("new_process"),
    recurse: bool = DefaultsToConfig("recurse"),
):
    """
        Makes selected targets; or all targets if none specified.

        Options:
            make recurse=1      Recursive make: put generated jobs in the
            queue.
            make new_process=1  Run the jobs in a new Python process.
            make echo=1         Displays the stdout/stderr for the job on
            the console.

            make new_process=1 echo=1   Not supported yet.
    """
    db = context.get_compmake_db()
    if not job_list:
        job_list = list(top_targets(db=db))

    manager = ManagerLocal(context=context, cq=cq, recurse=recurse, new_process=new_process, echo=echo)
    manager.add_targets(job_list)
    manager.process()
    return raise_error_if_manager_failed(manager)


@ui_command(section=ACTIONS, dbchange=True)
def pretend(
    job_list, context, cq,
):
    """
        Pretends that a target is done. The output is None.
    """
    db = context.get_compmake_db()
    if not job_list:
        job_list = list(top_targets(db=db))

    for job_id in job_list:
        i = IntervalTimer()
        i.stop()
        cache = Cache(Cache.DONE)
        cache.cputime_used = 0
        cache.walltime_used = 0
        cache.timestamp = time.time()
        cache.int_compute = cache.int_gc = i
        cache.int_load_results = cache.int_make = cache.int_save_results = i
        set_job_cache(job_id, cache, db)
        set_job_userobject(job_id, None, db)


@ui_command(section=ACTIONS, dbchange=True)
def invalidate(non_empty_job_list, context):
    """ Invalidates the cache of a job so that it will be remade. """
    db = context.get_compmake_db()
    for job in non_empty_job_list:
        mark_to_remake(job, db=db)


@ui_command(section=ACTIONS, dbchange=True)
def remake(
    non_empty_job_list,
    context,
    cq,
    echo: bool = DefaultsToConfig("echo"),
    new_process: bool = DefaultsToConfig("new_process"),
    recurse: bool = DefaultsToConfig("recurse"),
):
    """
        Remake the selected targets (equivalent to clean and make).

        :param non_empty_job_list:
        :param context:
        :param cq:
        :param echo:
        :param new_process:Run the jobs in a new Python process.
        :param recurse:   Recursive remake: put generated jobs in
    """
    non_empty_job_list = list(non_empty_job_list)

    if not ask_if_sure_remake(non_empty_job_list):
        return

    db = context.get_compmake_db()
    for job in non_empty_job_list:
        mark_to_remake(job, db=db)

    manager = ManagerLocal(context=context, cq=cq, recurse=recurse, new_process=new_process, echo=echo)

    manager.add_targets(non_empty_job_list)
    manager.process()
    return raise_error_if_manager_failed(manager)


@ui_command(section=ACTIONS, dbchange=True)
def rmake(
    job_list,
    context,
    cq,
    echo: bool = DefaultsToConfig("echo"),
    new_process: bool = DefaultsToConfig("new_process"),
):
    """ make with recurse = 1 """
    return make(job_list=job_list, context=context, cq=cq, echo=echo, new_process=new_process, recurse=True)



from nose.tools import istest
from compmake.unittests.compmake_test import CompmakeTest
from compmake.unittests.mockup_dynamic_4 import mockup_dyn4


@istest
class TestDynamic4(CompmakeTest):
    def test_dynamic4(self):
        mockup_dyn4(self.cc)
        self.assert_cmd_success("ls")
        self.assert_cmd_success("parmake recurse=1 echo=1")
        self.assert_cmd_success("ls")
        self.assert_cmd_success("details schedule")
        self.assert_cmd_success("details report")
        self.assert_cmd_success("clean")
        self.assert_cmd_success("ls")
        self.assert_cmd_success("parmake recurse=1 echo=1")

    def test_dynamic4b(self):
        mockup_dyn4(self.cc)
        self.assert_cmd_success("ls")
        self.assert_cmd_success("parmake recurse=1 echo=1")
        self.assert_cmd_success("ls")
        self.assert_cmd_success("details schedule")
        self.assert_cmd_success("details report")
        self.assert_cmd_success("clean schedule")
        self.assert_cmd_success("ls")
        self.assert_cmd_success("parmake recurse=1 echo=1")

    def test_dynamic4c(self):
        mockup_dyn4(self.cc)
        self.assert_cmd_success("ls")
        self.assert_cmd_success("parmake recurse=1 echo=1")
        self.assert_cmd_success("ls")
        self.assert_cmd_success("details schedule")
        self.assert_cmd_success("details report")
        self.assert_cmd_success("clean report")
        self.assert_cmd_success("ls")
        self.assert_cmd_success("parmake recurse=1 echo=1")



from nose.tools import istest

from .compmake_test import CompmakeTest
from compmake.unittests.mockup import mockup_recursive_5


@istest
class TestDynamic2(CompmakeTest):
    def test_dynamic2(self):
        mockup_recursive_5(self.cc)
        for _ in range(5):
            self.assert_cmd_success("ls;make;ls")
        self.assertJobsEqual("all", ["r1", "r2", "r3", "r4", "r5"])
        self.assertJobsEqual("done", ["r1", "r2", "r3", "r4", "r5"])

        self.assert_cmd_success("details r5")
        self.assert_cmd_success("details r4")
        self.assert_cmd_success("details r3")
        self.assert_cmd_success("details r2")
        self.assert_cmd_success("details r1")

        self.assert_defined_by("r5", ["root"])
        self.assert_defined_by("r4", ["root", "r5"])
        self.assert_defined_by("r3", ["root", "r5", "r4"])
        self.assert_defined_by("r2", ["root", "r5", "r4", "r3"])
        self.assert_defined_by("r1", ["root", "r5", "r4", "r3", "r2"])



# Jan 14 copied from conf_tools
import traceback

#
# from contracts import contract
# from contracts.utils import indent


__all__ = [
    "import_name",
]


# @contract(name="unicode")
from zuper_commons.text import indent


def import_name(name: str):
    """
        Loads the python object with the given name.

        Note that "name" might be "module.module.name" as well.
    """
    try:
        return __import__(name, fromlist=["dummy"])
    except ImportError as e:
        # split in (module, name) if we can
        if "." in name:
            tokens = name.split(".")
            field = tokens[-1]
            module_name = ".".join(tokens[:-1])

            if False:  # previous method
                pass
                # try:
                #     module = __import__(module_name, fromlist=['dummy'])
                # except ImportError as e:
                #     msg = ('Cannot load %r (tried also with %r):\n' %
                #            (name, module_name))
                #     msg += '\n' + indent(
                #         '%s\n%s' % (e, traceback.format_exc(e)), '> ')
                #     raise ValueError(msg)
                #
                # if not field in module.__dict__:
                #     msg = 'No field  %r\n' % field
                #     msg += ' found in %r.' % module
                #     raise ValueError(msg)
                #
                # return module.__dict__[field]
            else:
                # other method, don't assume that in "M.x", "M" is a module.
                # It could be a class as well, and "x" be a staticmethod.
                try:
                    module = import_name(module_name)
                except ImportError as e:
                    msg = "Cannot load %r (tried also with %r):\n" % (name, module_name)
                    msg += "\n" + indent("%s\n%s" % (e, traceback.format_exc()), "> ")
                    raise ValueError(msg)

                if not field in module.__dict__:
                    msg = "No field %r\n" % field
                    msg += " found in %r." % module
                    raise ValueError(msg)

                f = module.__dict__[field]

                # "staticmethod" are not functions but descriptors, we need
                # extra magic
                if isinstance(f, staticmethod):
                    return f.__get__(module, None)
                else:
                    return f

        else:
            msg = "Cannot import name %r." % name
            msg += indent(traceback.format_exc(), "> ")
            raise ValueError(msg)



# ssh -R 12000:localhost:6379 nessa.cds.caltech.edu
# "compmake --db=redis --host localhost:12000 --slave make
# v_rangefinder_nonunif-random_pose_simulation "
from collections import namedtuple
from pprint import pprint
import sys


Host = namedtuple("Host", "name host username processors init test instance")


def fill_in(config, defaults):
    for k, v in defaults.items():
        config[k] = config.get(k, v)


def parse_yaml_configuration(file):  # @ReservedAssignment
    import yaml

    configuration = yaml.load(file)

    pprint(configuration)

    results = {}

    types = configuration.get("types", {})
    hosts = configuration["hosts"]

    default_conf = {
        "processors": 1,
        "init": None,
        "test": None,
        "username": None,
        "host": None,
        "instance": 0,
    }
    default_type = types.get("default", {})
    fill_in(default_type, default_conf)
    types["default"] = default_type

    for compname, config in types.items():
        fill_in(config, default_type)

    for hostname, config in hosts.items():
        if not "host" in config:
            config["host"] = hostname
        config["name"] = hostname

        comptype = config.get("type", "default")
        assert comptype in types

        fill_in(config, types[comptype])

        assert not hostname in results, "Duplicated key"

        pprint(config)
        results[hostname] = Host(**config)

    return results


if __name__ == "__main__":
    hosts_config = parse_yaml_configuration(sys.stdin)
    print(hosts_config)



"""
    These are all wrappers around the raw methods in storage
"""
from typing import Iterator

from compmake.exceptions import CompmakeBug, CompmakeDBError, CompmakeException
from compmake.utils.pickle_frustration import pickle_main_context_load
from zuper_commons.types import check_isinstance, raise_desc

from ..structures import Cache, CMJobID, DBKey, Job
from ..utils import wildcard_to_regexp

KEY_JOB_PREFIX = "cm-job-"


def job2key(job_id: CMJobID) -> DBKey:
    return DBKey(f"{KEY_JOB_PREFIX}{job_id}")


def key2job(key: DBKey) -> CMJobID:
    return CMJobID(key.replace(KEY_JOB_PREFIX, "", 1))


def all_jobs(db, force_db: bool = False) -> Iterator[CMJobID]:
    """ Returns the list of all jobs.
        If force_db is True, read jobs from DB.
        Otherwise, use local cache.
     """
    pattern = job2key(CMJobID("*"))
    regexp = wildcard_to_regexp(pattern)

    for key in db.keys():
        if regexp.match(key):
            yield key2job(key)


def get_job(job_id: CMJobID, db) -> Job:
    key = job2key(job_id)
    computation = db[key]
    assert isinstance(computation, Job)
    return computation


def job_exists(job_id: CMJobID, db) -> bool:
    key = job2key(job_id)
    return key in db


def assert_job_exists(job_id: CMJobID, db):
    """
        :raise CompmakeBug: if the job does not exist
    """
    get_job(job_id, db)


def set_job(job_id: CMJobID, job: Job, db) -> None:
    # TODO: check if they changed
    key = job2key(job_id)
    assert isinstance(job, Job)
    db[key] = job


def delete_job(job_id: CMJobID, db) -> None:
    key = job2key(job_id)
    del db[key]


#
# Cache objects
#
def job2cachekey(job_id: CMJobID) -> str:
    prefix = "cm-cache-"
    return f"{prefix}{job_id}"


def get_job_cache(job_id, db):
    cache_key = job2cachekey(job_id)
    if cache_key in db:
        try:
            cache = db[cache_key]
            assert isinstance(cache, Cache)
        except Exception as e:
            del db[cache_key]
            # also remove user object?
            msg = f'Could not read Cache object for job "{job_id}": {e}; deleted.'
            raise CompmakeException(msg)
        return cache
    else:
        # make sure this is a valid job_id
        # XXX expensive
        # known = all_jobs()
        # if not job_id in known:
        # raise CompmakeException("invalid job %s, I know %s"
        # % (job_id, known))
        if not job_exists(job_id, db):
            raise_desc(CompmakeDBError, "Requesting cache for job that does not exist.", job_id=job_id)

        cache = Cache(Cache.NOT_STARTED)
        # we only put it later: NOT_STARTEd == not existent
        # get_compmake_db().set(cache_key, cache)
        return cache


def job_cache_exists(job_id: CMJobID, db) -> bool:
    key = job2cachekey(job_id)
    return key in db


def job_cache_sizeof(job_id: CMJobID, db) -> int:
    key = job2cachekey(job_id)
    return db.sizeof(key)


def set_job_cache(job_id: CMJobID, cache: Cache, db):
    assert isinstance(cache, Cache)
    check_isinstance(cache.captured_stderr, (type(None), str))
    check_isinstance(cache.captured_stdout, (type(None), str))
    check_isinstance(cache.exception, (type(None), str))
    check_isinstance(cache.backtrace, (type(None), str))
    key = job2cachekey(job_id)
    db[key] = cache


def delete_job_cache(job_id: CMJobID, db):
    key = job2cachekey(job_id)
    del db[key]


#
# User objects
#
def job2userobjectkey(job_id: CMJobID):
    prefix = "cm-res-"
    return f"{prefix}{job_id}"


def get_job_userobject(job_id: CMJobID, db) -> object:
    # available = is_job_userobject_available(job_id, db)
    # if not available:
    # available_job = job_exists(job_id, db)
    #         available_cache = job_cache_exists(job_id, db)
    #         msg = 'Job user object %r does not exist.' % job_id
    #         msg += ' Job exists: %s. Cache exists: %s. ' % (available_job,
    #  available_cache)
    #         msg += '\n jobs: %s' % list(all_jobs(db))
    #         msg += '\n path: %s' % db.basepath
    #         raise CompmakeBug(msg)
    # print('loading %r ' % job_id)
    key = job2userobjectkey(job_id)
    res = db[key]
    # print('... done')
    return res


def job_userobject_sizeof(job_id: CMJobID, db) -> int:
    key = job2userobjectkey(job_id)
    return db.sizeof(key)


def is_job_userobject_available(job_id: CMJobID, db) -> bool:
    key = job2userobjectkey(job_id)
    return key in db


job_userobject_exists = is_job_userobject_available


def set_job_userobject(job_id: CMJobID, obj, db):
    key = job2userobjectkey(job_id)
    db[key] = obj


def delete_job_userobject(job_id: CMJobID, db):
    key = job2userobjectkey(job_id)
    del db[key]


def job2jobargskey(job_id: CMJobID):
    prefix = "cm-args-"
    return f"{prefix}{job_id}"


def get_job_args(job_id: CMJobID, db):
    key = job2jobargskey(job_id)

    # if False:
    #     return db[key]
    # else:
    job = get_job(job_id, db)
    pickle_main_context = job.pickle_main_context
    with pickle_main_context_load(pickle_main_context):
        return db[key]


def job_args_exists(job_id: CMJobID, db) -> bool:
    key = job2jobargskey(job_id)
    return key in db


def job_args_sizeof(job_id: CMJobID, db) -> int:
    key = job2jobargskey(job_id)
    return db.sizeof(key)


def set_job_args(job_id: CMJobID, obj, db):
    key = job2jobargskey(job_id)
    db[key] = obj


def delete_job_args(job_id: CMJobID, db):
    key = job2jobargskey(job_id)
    del db[key]


def delete_all_job_data(job_id: CMJobID, db):
    # print('deleting_all_job_data(%r)' % job_id)
    args = dict(job_id=job_id, db=db)
    if job_exists(**args):
        delete_job(**args)
    if job_args_exists(**args):
        delete_job_args(**args)
    if job_userobject_exists(**args):
        delete_job_userobject(**args)
    if job_cache_exists(**args):
        delete_job_cache(**args)


# These are delicate and should be implemented differently
def db_job_add_dynamic_children(job_id: CMJobID, children, returned_by, db):
    job = get_job(job_id, db)
    if not returned_by in job.children:
        msg = f"{job_id!r} does not know it has child  {returned_by!r}"
        raise CompmakeBug(msg)

    job.children.update(children)
    job.dynamic_children[returned_by] = children
    set_job(job_id, job, db)
    job2 = get_job(job_id, db)
    assert job2.children == job.children, "Race condition"
    assert job2.dynamic_children == job.dynamic_children, "Race condition"


def db_job_add_parent(db, job_id, parent):
    j = get_job(job_id, db)
    # print('%s old parents list: %s' % (d, j.parents))
    j.parents.add(parent)
    set_job(job_id, j, db)
    j2 = get_job(job_id, db)
    assert j2.parents == j.parents, "Race condition"  # FIXME


def db_job_add_parent_relation(child: CMJobID, parent: CMJobID, db):
    child_comp = get_job(child, db=db)
    orig = set(child_comp.parents)
    want = orig | {parent}
    # alright, need to take care of race condition
    while True:
        # Try to write
        child_comp.parents = want
        set_job(child, child_comp, db=db)
        # now read back
        child_comp = get_job(child, db=db)
        if child_comp.parents != want:
            print(f"race condition for parents of {child}")
            print(f"orig: {orig}")
            print(f"want: {want}")
            print(f"now: {child_comp.parents}")
            # add the children of the other racers as well
            want = want | child_comp.parents
        else:
            break



import os

from .qacct import JobNotRunYet, get_qacct
from compmake.jobs import AsyncResultInterface, result_dict_raise_if_error
from compmake.exceptions import CompmakeBug, HostFailed

from compmake.utils import safe_pickle_load, which
from zuper_commons.text import indent
from system_cmd import CmdException, system_cmd_result


__all__ = [
    "SGESub",
]


class SGESub:
    def __init__(self, name, db, spool):
        self.name = name
        self.job_id = None
        self.db = db
        self.res = None
        self.spool = spool

    def delete_job(self):
        self.res.delete_job()

    def instance_job(self, job_id):
        self.res = SGEJob(job_id, self.db, spool=self.spool)
        return self.res


class SGEJob(AsyncResultInterface):
    compmake_bin = None

    @staticmethod
    def get_compmake_bin():
        """ Returns the path to the compmake executable. """
        if SGEJob.compmake_bin is None:
            compmake_bin = which("compmake")
            SGEJob.compmake_bin = compmake_bin
        return SGEJob.compmake_bin

    def __init__(self, job_id, db, spool):
        self.job_id = job_id
        self.db = db
        self.execute(spool=spool)

    def delete_job(self):
        # cmd = ['qdel', self.sge_id]
        cmd = ["qdel", self.sge_job_name]
        cwd = os.path.abspath(os.getcwd())
        # TODO: check errors
        try:
            _ = system_cmd_result(
                cwd,
                cmd,
                display_stdout=False,
                display_stderr=False,
                raise_on_error=True,
                capture_keyboard_interrupt=False,
            )
        except CmdException as e:
            print("Error while deleting job:\n%s" % e)

    def execute(self, spool):
        db = self.db
        # Todo: check its this one

        storage = os.path.abspath(db.basepath)

        # create a new spool directory for each execution
        # otherwise we get confused!

        self.stderr = os.path.join(spool, "%s.stderr" % self.job_id)
        self.stdout = os.path.join(spool, "%s.stdout" % self.job_id)
        self.retcode = os.path.join(spool, "%s.retcode" % self.job_id)
        self.out_results = os.path.join(spool, "%s.results.pickle" % self.job_id)

        if os.path.exists(self.stderr):
            os.remove(self.stderr)

        if os.path.exists(self.stdout):
            os.remove(self.stdout)

        if os.path.exists(self.retcode):
            os.remove(self.retcode)

        if os.path.exists(self.out_results):
            os.remove(self.out_results)

        options = []
        cwd = os.path.abspath(os.getcwd())
        variables = dict(SGE_O_WORKDIR=cwd, PYTHONPATH=os.getenv("PYTHONPATH", "") + ":" + cwd)

        # nice-looking name
        self.sge_job_name = "cm%s-%s" % (os.getpid(), self.job_id)
        # Note that we get the official "id" later and we store it in
        # self.sge_id
        options.extend(["-v", ",".join("%s=%s" % x for x in variables.items())])
        # XXX: spaces
        options.extend(["-e", self.stderr])
        options.extend(["-o", self.stdout])
        options.extend(["-N", self.sge_job_name])
        options.extend(["-wd", cwd])

        options.extend(["-V"])  # pass all environment

        options.extend(["-terse"])

        compmake_bin = SGEJob.get_compmake_bin()

        compmake_options = [
            compmake_bin,
            storage,
            "--retcodefile",
            self.retcode,
            "--status_line_enabled",
            "0",
            "--colorize",
            "0",
            "-c",
            '"make_single out_result=%s %s"' % (self.out_results, self.job_id),
        ]

        # if not all_disabled():
        #     compmake_options += ["--contracts"]

        # XXX: spaces in variable out_result

        write_stdin = " ".join(compmake_options)

        cmd = ["qsub"] + options

        res = system_cmd_result(
            cwd,
            cmd,
            display_stdout=False,
            display_stderr=True,
            raise_on_error=True,
            write_stdin=write_stdin,
            capture_keyboard_interrupt=False,
        )

        self.sge_id = res.stdout.strip()

        self.already_read = False
        self.npolls = 0

        self.told_you_ready = False

    def ready(self):
        if self.told_you_ready:
            raise CompmakeBug("should not call ready() twice")

        if self.npolls % 20 == 1:
            try:
                qacct = get_qacct(self.sge_id)
                # print('job: %s sgejob: %s res: %s' % (self.job_id,
                # self.sge_id, qacct))
                if "failed" in qacct and qacct["failed"] != "0":
                    reason = "Job schedule failed: %s\n%s" % (qacct["failed"], qacct)
                    raise HostFailed(host="xxx", job_id=self.job_id, reason=reason, bt="")  # XXX

            except JobNotRunYet:
                qacct = None
                pass
        else:
            qacct = None

        self.npolls += 1

        if os.path.exists(self.retcode):
            self.told_you_ready = True
            return True
        else:
            if qacct is not None:
                msg = "The file %r does not exist but it looks like the job " "is done" % self.retcode
                msg += "\n %s " % qacct
                # All right, this is simply NFS that is not updated yet
                # raise CompmakeBug(msg)

            return False

    def get(self, timeout=0):
        if not self.told_you_ready:
            raise CompmakeBug("I didnt tell you it was ready.")
        if self.already_read:
            msg = "Compmake BUG: should not call twice."
            raise CompmakeBug(msg)
        self.already_read = True

        assert os.path.exists(self.retcode)
        ret_str = open(self.retcode, "r").read()
        try:
            ret = int(ret_str)
        except ValueError:
            msg = "Could not interpret file %r: %r." % (self.retcode, ret_str)
            raise HostFailed(host="localhost", job_id=self.job_id, reason=msg, bt="")
            #
        #
        #         raise HostFailed(host="xxx",
        #                                      job_id=self.job_id,
        # reason=reason, bt="")  # XXX
        #

        try:
            stderr = open(self.stderr, "r").read()
            stdout = open(self.stdout, "r").read()

            stderr = "Contents of %s:\n" % self.stderr + stderr
            stdout = "Contents of %s:\n" % self.stdout + stdout

            # if ret == CompmakeConstants.RET_CODE_JOB_FAILED:
            #                 msg = 'SGE Job failed (ret: %s)\n' % ret
            #                 msg += indent(stderr, '| ')
            #                 # mark_as_failed(self.job_id, msg, None)
            #                 raise JobFailed(msg)
            #             elif ret != 0:
            #                 msg = 'SGE Job failed (ret: %s)\n' % ret
            #                 error(msg)
            #                 msg += indent(stderr, '| ')
            #                 raise JobFailed(msg)

            if not os.path.exists(self.out_results):
                msg = "job succeeded but no %r found" % self.out_results
                msg += "\n" + indent(stderr, "stderr")
                msg += "\n" + indent(stdout, "stdout")
                raise CompmakeBug(msg)

            res = safe_pickle_load(self.out_results)
            result_dict_raise_if_error(res)
            return res
        finally:
            fs = [self.stderr, self.stdout, self.out_results, self.retcode]
            for filename in fs:
                if os.path.exists(filename):
                    os.unlink(filename)



# Code copied from PEP-0257

__all__ = ["docstring_trim", "docstring_components"]


def docstring_trim(docstring):
    if not docstring:
        return ""
    # Convert tabs to spaces (following the normal Python rules)
    # and split into a list of lines:
    lines = docstring.expandtabs().splitlines()
    # Determine minimum indentation (first line doesn't count):
    indent = 10000
    for line in lines[1:]:
        stripped = line.lstrip()
        if stripped:
            indent = min(indent, len(line) - len(stripped))
    # Remove indentation (first line is special):
    trimmed = [lines[0].strip()]
    if indent < 10000:
        for line in lines[1:]:
            trimmed.append(line[indent:].rstrip())
    # Strip off trailing and leading blank lines:
    while trimmed and not trimmed[-1]:
        trimmed.pop()
    while trimmed and not trimmed[0]:
        trimmed.pop(0)
    # Return a single string:
    return "\n".join(trimmed)


def docstring_components(docstring):
    """

        Removes leading whitespace and returns a dict
        with fields "first" and "rest".

        This is the rest.

    """

    # first, remove whitespace
    docstring = docstring_trim(docstring)
    # split in newlines
    lines = docstring.split("\n")
    # trim each one
    lines = list(map(lambda x: x.strip(), lines))
    # remove initial empty ones
    while lines and lines[0] == "":
        lines.pop(0)
    first = []
    while lines and lines[0] != "":
        first.append(lines.pop(0))

    # remove separation
    while lines and lines[0] == "":
        lines.pop(0)
    rest = lines

    res = {"first": " ".join(first), "rest": "\n".join(rest)}
    # remove newlines here
    return res


def docstring_components_test():
    res = docstring_components(docstring_components.__doc__)
    print(res)

    assert res["first"] == "Removes leading whitespace and returns a dict " 'with fields "first" and "rest".'
    assert res["rest"] == "This is the rest."



import os
import sys
import traceback

from asciimatics.exceptions import StopApplication
from asciimatics.scene import Scene
from asciimatics.screen import Screen
from asciimatics.widgets import Frame, Layout, TextBox
from future import builtins

from compmake import get_compmake_config, logger
from zuper_commons.text import indent, remove_escapes
from .ui import clean_other_jobs, get_commands, interpret_commands
from .visualization import clean_console_line, DefaultConsole, ui_error
from .. import CompmakeConstants, CompmakeGlobalState, set_compmake_status
from ..events import Event, publish, register_handler
from ..exceptions import CommandFailed, CompmakeBug, JobInterrupted, MakeFailed, ShellExitRequested, UserError
from ..jobs import all_jobs, CacheQueryDB

__all__ = [
    "interactive_console",
    "interpret_commands_wrap",
    "batch_command",
    "compmake_console_text",
    "compmake_console_gui",
]


def get_readline():
    """
    Returns a reference to the readline (or Pyreadline) module if they
    are available and the config "readline" is True, otherwise None.
    :return:Reference to readline module or None
    """
    use_readline = get_compmake_config("readline")
    if not use_readline:
        return None
    else:
        try:
            import readline

            return readline
        except BaseException as e:
            try:
                # noinspection PyUnresolvedReferences
                import pyreadline as readline  # @UnresolvedImport

                return readline
            except Exception as e2:
                # TODO: write message
                msg = "Neither readline or pyreadline available."
                msg += "\n- readline error: %s" % e
                msg += "\n- pyreadline error: %s" % e2
                logger.warning(msg)
                return None


# @contract(cq=CacheQueryDB, returns="None")
def interpret_commands_wrap(commands, context, cq: CacheQueryDB) -> None:
    """
        Returns None or raises CommandFailed, ShellExitRequested,
            CompmakeBug, KeyboardInterrupt.
    """
    assert context is not None
    publish(context, "command-line-starting", command=commands)

    try:
        interpret_commands(commands, context=context, cq=cq)
        publish(context, "command-line-succeeded", command=commands)
    except CompmakeBug:
        raise
    except UserError as e:
        publish(context, "command-line-failed", command=commands, reason=e)
        raise CommandFailed(str(e)) from e
    except CommandFailed as e:
        publish(context, "command-line-failed", command=commands, reason=e)
        raise
    except (KeyboardInterrupt, JobInterrupted) as e:
        publish(context, "command-line-interrupted", command=commands, reason="KeyboardInterrupt")
        # If debugging
        # tb = traceback.format_exc()
        # print tb  # XXX
        raise CommandFailed(str(e)) from e
        # raise CommandFailed('Execution of %r interrupted.' % commands)
    except ShellExitRequested:
        raise
    except Exception as e:
        tb = traceback.format_exc()
        msg0 = (
            "Warning, I got this exception, while it should "
            "have been filtered out already. "
            "This is a compmake BUG that should be reported "
            "at http://github.com/AndreaCensi/compmake/issues"
        )
        msg = msg0 + "\n" + indent(tb, "bug| ")
        publish(context, "compmake-bug", user_msg=msg, dev_msg="")  # XXX
        raise CompmakeBug(msg) from e


def interactive_console(context):
    """
        raises: CommandFailed, CompmakeBug
    """
    publish(context, "console-starting")

    # shared cache query db by commands
    cq = CacheQueryDB(context.get_compmake_db())

    while True:
        try:
            for line in compmake_console_lines(context):
                interpret_commands_wrap(line, context=context, cq=cq)
        except CommandFailed as e:
            if not isinstance(e, MakeFailed):
                ui_error(context, str(e))
            continue
        except CompmakeBug:
            raise
        except ShellExitRequested:
            break
        except KeyboardInterrupt:  # CTRL-C
            print("\nPlease use 'exit' to quit.")
        except EOFError:  # CTRL-D
            # TODO maybe make loop different? we don't want to catch
            # EOFerror in interpret_commands
            print("(end of input detected)")
            break

    publish(context, "console-ending")
    return None


def get_completions(context):
    db = context.get_compmake_db()
    if CompmakeGlobalState.cached_completions is None:
        available = get_commands().keys()
        available.extend(list(all_jobs(db=db)))  # give it a list
        # TODO: add function type "myfunc()"
        CompmakeGlobalState.cached_completions = available

    return CompmakeGlobalState.cached_completions


def tab_completion2(context, text, state):
    completions = get_completions(context=context)
    matches = sorted(x for x in completions if x.startswith(text))
    try:
        response = matches[state]
    except IndexError:
        response = None
    return response


# TODO: move
COMPMAKE_HISTORY_FILENAME = ".compmake_history.txt"


def compmake_console_lines(context):
    """ Returns lines with at least one character. """
    readline = get_readline()

    if readline is not None:
        try:
            # Rewrite history
            # TODO: use readline's support for history
            if os.path.exists(COMPMAKE_HISTORY_FILENAME):
                with open(COMPMAKE_HISTORY_FILENAME) as f:
                    lines = f.read().split("\n")

                with open(COMPMAKE_HISTORY_FILENAME, "w") as f:
                    last_word = None
                    for word in lines:
                        word = word.strip()
                        if len(word) == 1:
                            continue  # 'y', 'n'
                        if word in ["exit", "quit", "ls"]:
                            continue
                        if word == last_word:  # no doubles
                            continue
                        f.write("%s\n" % word)
                        last_word = word

            # noinspection PyUnresolvedReferences
            readline.read_history_file(COMPMAKE_HISTORY_FILENAME)
        except:
            pass

        # noinspection PyUnresolvedReferences
        readline.set_history_length(300)
        completer = lambda text, state: tab_completion2(context, text, state)
        # noinspection PyUnresolvedReferences
        readline.set_completer(completer)
        # noinspection PyUnresolvedReferences
        readline.set_completer_delims(" ")
        # noinspection PyUnresolvedReferences
        readline.parse_and_bind("tab: complete")

    while True:
        clean_console_line(sys.stdout)

        # TODO: find alternative, not reliable if colored
        # line = raw_input(colored('@: ', 'cyan'))
        line = builtins.input("@: ")
        line = line.strip()
        if not line:
            continue

        if readline is not None:
            # noinspection PyUnresolvedReferences
            readline.write_history_file(COMPMAKE_HISTORY_FILENAME)

        yield line


# noinspection PyUnresolvedReferences
def ask_question(question, allowed=None):
    """ Asks a yes/no question to the user """
    readline = get_readline()
    if allowed is None:
        allowed = {"y": True, "Y": True, "yes": True, "n": False, "N": False, "no": False}
    while True:
        line = builtins.input(question)
        line = line.strip()

        # we don't want these to go into the history
        if readline is not None:
            try:
                l = readline.get_current_history_length()
                if l:
                    readline.remove_history_item(l - 1)
            except:
                pass

        if line in allowed:
            return allowed[line]


# Note: we wrap these in shallow functions because we don't want
# to import other things.


def batch_command(s, context, cq):
    """
        Executes one command (could be a sequence)

        Returns None or raises CommandsFailed, CompmakeBug.
    """

    set_compmake_status(CompmakeConstants.compmake_status_embedded)

    # we assume that we are done with defining jobs
    clean_other_jobs(context=context)
    from compmake.scripts.master import read_rc_files

    read_rc_files(context=context)
    return interpret_commands_wrap(s, context=context, cq=cq)


def compmake_console_text(context):
    clean_other_jobs(context=context)
    from compmake.scripts.master import read_rc_files

    read_rc_files(context=context)
    interactive_console(context=context)


def compmake_console_gui(context):
    return Screen.wrapper(compmake_console_gui_, arguments=[context])


def compmake_console_gui_(screen: Screen, context):
    clean_other_jobs(context=context)
    from compmake.scripts.master import read_rc_files

    read_rc_files(context=context)

    DefaultConsole.active = False
    scenes = [
        Scene([MainList(screen, context)], -1, name="Main"),
    ]

    screen.play(scenes, stop_on_resize=False, start_scene=scenes[0], allow_int=True)


class MainList(Frame):
    offset: int

    def __init__(self, screen: Screen, context):
        self.all_lines = []
        self.offset = 0
        super(MainList, self).__init__(
            screen,
            screen.height,
            screen.width,
            on_load=self.on_load,
            hover_focus=True,
            can_scroll=False,
            title="Compmake",
        )

        # Save off the model that accesses the contacts database.
        self.context = context
        layout = Layout([100])
        self.add_layout(layout)
        cmd = None
        tb = TextBox(TextBox.FILL_FRAME, name="text", disabled=True, as_string=False,)
        tb.value = ["This is the ", "initial"]

        cq = CacheQueryDB(context.get_compmake_db())

        def on_command_change(*args):
            v = cmd.value
            # print(repr(v))
            # tb.value = [repr(v)]
            if len(v) > 1:
                self.save()
                line = v[0]
                cmd.value = [""]
                screen.refresh()

                try:
                    interpret_commands_wrap(line, context=context, cq=cq)
                except ShellExitRequested:
                    raise StopApplication("bye")
                except CommandFailed as e:
                    ui_error(context, str(e))

            # if v.endswith('\n'):
            #     tb.value += v
            # cmd.value = ''

        cmd = TextBox(1, "command", on_change=on_command_change)
        status = TextBox(1, "status", disabled=True)
        layout.add_widget(tb, 0)
        layout.add_widget(status, 0)
        layout.add_widget(cmd, 0)

        H = 25

        def handle_ui_message(context, event: Event):
            string = event.kwargs["string"]
            string = remove_escapes(string)
            self.all_lines = self.all_lines + string.split("\n")
            self.offset = len(self.all_lines) - H
            tb.value = self.all_lines[self.offset : self.offset + 25]
            screen.refresh()

        register_handler("ui-message", handle_ui_message)

        def handle_ui_status_summary(context, event: Event):
            string = event.kwargs["string"]
            string = remove_escapes(string)
            status.value = [string]
            screen.refresh()

        register_handler("ui-status-summary", handle_ui_status_summary)

        self.fix()

    def on_load(self, new_value=None):
        pass



from nose.tools import istest

from .compmake_test import CompmakeTest
from .test_dynamic_1 import mockup_dynamic1, TestDynamic1


@istest
class TestDynamic1rec(CompmakeTest):

    howmany = None  # used by cases()

    def test_dynamic1(self):
        TestDynamic1.howmany = 3
        mockup_dynamic1(self.cc)
        # At this point we have generated only two jobs
        self.assertJobsEqual("all", ["generate", "values"])

        # now we make them
        self.assert_cmd_success("make recurse=1")

        # this will have created new jobs
        self.assertJobsEqual(
            "all", ["generate", "values", "actual0", "actual1", "actual2", "generate-finish"]
        )
        # ... still to do
        self.assertJobsEqual(
            "done", ["generate", "values", "actual0", "actual1", "actual2", "generate-finish"]
        )



import functools
import nose


def expected_failure(test):
    @functools.wraps(test)
    def inner(*args, **kwargs):
        try:
            test(*args, **kwargs)
        except Exception:
            raise nose.SkipTest
        else:
            raise AssertionError("Failure expected")

    return inner



import sys

__all__ = [
    "termcolor_colored",
]

# Available text colors:
#         red, green, yellow, blue, magenta, cyan, white.
#
#     Available text highlights:
#         on_red, on_green, on_yellow, on_blue, on_magenta, on_cyan, on_white.
#
#     Available attributes:
#         bold, dark, underline, blink, reverse, concealed.
#
#

from termcolor import colored as t_colored  # @UnresolvedImport

from zuper_commons.types import check_isinstance


def termcolor_colored(s: str, color=None, on_color=None, attrs=None):
    check_isinstance(s, str)
    return "\n".join(t_colored(x, color, on_color, attrs) for x in s.split("\n"))


#
# try:
#
#
# except:
#     # TODO: logger
#     sys.stderr.write('compmake can make use of the package "termcolor".' " Please install it.\n")
#
#     def termcolor_colored(x, color=None, on_color=None, attrs=None):
#         """ emulation of the termcolor interface """
#         return x



from contextlib import contextmanager
from zuper_commons.text import indent
from system_cmd import CmdException, system_cmd_result
import os
import tempfile


def get_examples_path():
    from pkg_resources import resource_filename  # @UnresolvedImport

    here = resource_filename("compmake", "unittests")
    examples = os.path.join(here, "..", "examples")
    examples = os.path.abspath(examples)
    if not os.path.exists(examples):
        msg = "Example dir does not exist: %s" % examples
        raise Exception(msg)
    return examples


def run_example(name, command, expect_fail=False):
    examples = get_examples_path()
    pyfile = os.path.join(examples, "%s.py" % name)
    if not os.path.exists(pyfile):
        msg = "Example file does not exist: %s" % pyfile
        raise Exception(msg)

    with create_tmp_dir() as cwd:
        cmd = [pyfile, command]
        try:
            res = system_cmd_result(cwd, cmd, display_stdout=False, display_stderr=False, raise_on_error=True)
            if expect_fail:
                msg = "Expected failure of %s but everything OK." % name
                msg += "\n cwd = %s" % cwd
                msg += "\n" + indent(res.stderr, "stderr| ")
                msg += "\n" + indent(res.stdout, "stdout| ")
                raise Exception(msg)
            return res
        except CmdException as e:
            stderr = e.res.stderr
            stdout = e.res.stdout
            if not expect_fail:
                msg = "Example %r: Command %r failed unexpectedly." % (name, command)
                msg += "\n retcode: %r" % e.res.ret
                msg += "\n" + indent(stderr, "stderr| ")
                msg += "\n" + indent(stdout, "stdout| ")
                raise Exception(msg) from e


@contextmanager
def create_tmp_dir():
    # FIXME: does not delete dir
    dirname = tempfile.mkdtemp()
    try:
        yield dirname
    except:
        raise


cmd_make1 = "make recurse=1"
cmd_make2 = "parmake recurse=1"
cmd_make3 = "make recurse=1 new_process=1"
cmd_make4 = "parmake recurse=1 new_process=1"

# This gets slow
# def test_example_big1():
#     run_example('example_big', cmd_make1, expect_fail=True)
#
# def test_example_big2():
#     run_example('example_big', cmd_make2, expect_fail=True)
# def test_example_big3():
#     run_example('example_big', cmd_make3, expect_fail=True)
#
# def test_example_big4():
#     run_example('example_big', cmd_make4, expect_fail=True)


def test_example_dynamic_explicitcontext1():
    run_example("example_dynamic_explicitcontext", cmd_make1)


def test_example_dynamic_explicitcontext2():
    run_example("example_dynamic_explicitcontext", cmd_make2)


def test_example_progress1():
    run_example("example_progress", cmd_make1)


def test_example_progress2():
    run_example("example_progress", cmd_make2)


def test_example_progress_same1():
    run_example("example_progress_same", cmd_make1)


def test_example_progress_same2():
    run_example("example_progress_same", cmd_make2)


def test_example_progress_same3():
    run_example("example_progress_same", cmd_make3)


def test_example_progress_same4():
    run_example("example_progress_same", cmd_make4)


def test_example_simple1():
    run_example("example_simple", cmd_make1)


def test_example_simple2():
    run_example("example_simple", cmd_make2)


def example_external_support1():
    run_example("example_external_support", cmd_make1)


def example_external_support2():
    run_example("example_external_support", cmd_make2)


def example_external_support3():
    run_example("example_external_support", cmd_make3)


def example_external_support4():
    run_example("example_external_support", cmd_make4)


if True:

    # Fails for pickle reasons
    #     @expected_failure
    def test_example_dynamic_explicitcontext3():
        run_example("example_dynamic_explicitcontext", cmd_make3)

    #     @expected_failure
    def test_example_dynamic_explicitcontext4():
        run_example("example_dynamic_explicitcontext", cmd_make4)

    #     @expected_failure
    def test_example_progress3():
        run_example("example_progress", cmd_make3)

    #     @expected_failure
    def test_example_progress4():
        run_example("example_progress", cmd_make4)

    #     @expected_failure
    def test_example_simple3():
        run_example("example_simple", cmd_make3)

    #     @expected_failure
    def test_example_simple4():
        run_example("example_simple", cmd_make4)



from compmake.events.registrar import register_handler
from compmake.plugins.graph import graph
from compmake.ui.helpers import COMMANDS_ADVANCED, ui_command
from zuper_commons.fs import make_sure_dir_exists
from system_cmd.meat import system_cmd_result
import os


class Global:
    step = 0
    job_list = []
    graph_params = {}
    dirname = "."

    size = (None, None)
    dpi = None
    processing = set()


def update_graph(context, event):
    print("event: %s" % event)

    if event.name in ["manager-job-processing"]:
        job_id = event.kwargs["job_id"]
        Global.processing.add(job_id)
    if event.name in ["manager-job-failed", "manager-job-done"]:
        job_id = event.kwargs["job_id"]
        Global.processing.remove(job_id)

    print(f"global processing {Global.processing}")
    if "job_id" in event.kwargs:
        what = f"{event.name}-{event.kwargs['job_id']}"
    else:
        what = event.name
    filename = os.path.join(Global.dirname, f"step-{Global.step:04d}-{what}")

    make_sure_dir_exists(filename)
    #     print('step %d: jobs = %s' % (Global.step, Global.job_list))
    graph(
        job_list=list(Global.job_list),
        context=context,
        filename=filename,
        processing=Global.processing,
        **Global.graph_params,
    )
    Global.step += 1

    # see here:
    # http://stackoverflow.com/questions/14784405/how-to-set-the-output-size-in-graphviz-for-the-dot-format

    png = filename + ".png"
    png2 = filename + "-x.png"

    size = Global.size
    dpi = Global.dpi
    # noinspection PyUnresolvedReferences
    cmd0 = [
        "dot",
        "-Tpng",
        "-Gsize=%s,%s\!" % (size[0] / dpi, size[1] / dpi),
        "-Gdpi=%s" % dpi,
        "-o" + png,
        filename,
    ]
    system_cmd_result(".", cmd0, display_stdout=True, display_stderr=True, raise_on_error=True)

    cmd = [
        "convert",
        png,
        "-gravity",
        "center",
        "-background",
        "white",
        "-extent",
        "%sx%s" % (size[0], size[1]),
        png2,
    ]
    system_cmd_result(".", cmd, display_stdout=True, display_stderr=True, raise_on_error=True)
    os.unlink(png)


@ui_command(section=COMMANDS_ADVANCED, alias="graph-animation")
def graph_animation(
    job_list, context, dirname="compmake-graph-animation", dpi=150, width=900, height=900, label="function"
):
    """
        Runs a step-by-step animation.

        Registers the handlers. Then call 'make' or 'parmake'.
    """
    possible = ["none", "id", "function"]
    if not label in possible:
        msg = "Invalid label method %r not in %r." % (label, possible)
        raise ValueError(msg)

    Global.dirname = dirname
    Global.job_list = list(job_list)

    Global.graph_params = dict(filter="dot", format="png", label=label, color=True, cluster=True)
    Global.dpi = dpi
    Global.size = (width, height)
    Global.processing = set()
    events = [
        "manager-job-processing",
        "manager-job-failed",
        "manager-job-done",
        "manager-succeeded",
        "manager-phase",
    ]

    for e in events:
        register_handler(e, update_graph)



# -*- coding: utf-8 -*-

from compmake.jobs.storage import job2cachekey, job2jobargskey, job2key, job2userobjectkey
from compmake.jobs.uptodate import CacheQueryDB
from compmake.storage.filesystem import StorageFilesystem

import os

__all__ = ["synchronize_db_up"]


def synchronize_db_up(context, targets):
    """ Syncrhonizes the DB up """
    db = context.get_compmake_db()
    # first create the volume if it doesn't exist
    vol = create_db_volume(db)

    # now we need to put all files

    keys = []

    cq = CacheQueryDB(db)
    jobs = set()
    jobs.update(targets)
    jobs.update(cq.tree(targets))

    # print('%d targets, %d jobs' % (len(targets), len(jobs)))

    # XXX: not all jobs
    for job_id in jobs:
        resources = [job2jobargskey, job2userobjectkey, job2cachekey, job2key]
        for r in resources:
            key = r(job_id)
            if key in db:
                keys.append(key)

    # print('Found %s files to upload' % len(keys))

    # Shadow storage
    db2 = StorageFilesystem(basepath=vol.mount_path)
    already = set([os.path.basename(x["path"]) for x in vol.ls(".")])

    filename2contents = {}
    # print('obtained: %r' % already)
    for key in keys:
        f = db.filename_for_key(key)
        f2 = db2.filename_for_key(key)
        local_path = f
        remote_path = os.path.relpath(f2, db2.basepath)

        if remote_path in already:
            # print('skipping %r' % local_path)
            continue

        size = os.stat(local_path).st_size
        use_compact = size < 6 * 1024
        if use_compact:
            with open(local_path) as f:
                filename2contents[f2] = f.read()
        else:
            # print('%s -> %s' % (local_path, remote_path))
            assert os.path.join(db2.basepath, remote_path) == f2
            vol.put_file(local_path, remote_path, target_mode=None)

    import multyvac

    multyvac_job_id = multyvac.submit(copy_files, filename2contents, _vol=[vol.name])
    multyvac_job = multyvac.get(multyvac_job_id)
    multyvac_job.get_result()

    return vol, db2


def copy_files(filename2contents):
    s = ""
    for filename, contents in filename2contents.items():
        real = os.path.realpath(filename)
        with open(filename, "w") as f:
            f.write(contents)
        s += "%s = %s" % (filename, real) + "\n"
    return s


def delete_db_volume(db):
    vol = create_db_volume(db)
    entries = vol.ls(".")
    if not entries:
        return

    entries = [os.path.join(vol.mount_path, x["path"]) for x in entries]

    import multyvac

    multyvac_job_id = multyvac.submit(delete_entries, entries, _vol=[vol.name], _name="Reset Compmake DB")
    multyvac_job = multyvac.get(multyvac_job_id)
    multyvac_job.get_result()


def delete_entries(entries):
    print("Removing %d entries..." % len(entries))
    for x in entries:
        os.unlink(x)


def create_db_volume(db):
    """ Returns a Volume """
    name = os.path.basename(db.basepath)
    # print('using volume name %r' % name)
    mount_path = "/compmake-db"
    return get_or_create_volume(name, mount_path)


def get_or_create_volume(volume_name, mount_path):
    volumes = multyvac.volume.list()  # @UndefinedVariable

    for vol in volumes:
        if vol.name == volume_name:
            return vol
    else:
        multyvac.volume.create(volume_name, mount_path)  # @UndefinedVariable
        vol = multyvac.volume.get(volume_name)  # @UndefinedVariable
        return vol


def synchronize_dir(syncdir):
    if not os.path.exists(syncdir):
        msg = "Dir %r does not exist." % syncdir
        raise ValueError(msg)

    import multyvac

    syncdir = os.path.realpath(syncdir)

    if not syncdir or not syncdir[0] == "/":
        msg = "I expect the multyvac_sync dir to be an absolute path,"
        msg += " got %r." % syncdir
        raise ValueError(msg)

    # local='/data/work/datasets/00-ldr21-logs/'
    # Let's use the first path
    tokens = syncdir.split("/")
    if tokens < 3:
        msg = "Got %r." % str(tokens)
        raise ValueError(msg)

    volume_name = tokens[1]
    mount_path = "/" + volume_name
    rest = "/".join(tokens[2:])

    volumes = [v.name for v in multyvac.volume.list()]  # @UndefinedVariable

    # print('obtained list of volumes %r' % volumes)
    if not volume_name in volumes:
        # print('creating new volume')
        multyvac.volume.create(volume_name, mount_path)  # @UndefinedVariable

    vol = multyvac.volume.get(volume_name)  # @UndefinedVariable

    vol.mkdir(rest)
    rest_minus = "/".join(tokens[2:-1])
    # print('synchronizing up - sync_up(%r,%r)' % (syncdir, rest_minus))

    vol.sync_up(syncdir, rest_minus)
    # print('synchronizing up done.')

    return vol.name



# -*- coding: utf-8 -*-

from .mvac_manager import MVacManager
from compmake.constants import DefaultsToConfig
from compmake.events import publish
from compmake.jobs import top_targets
from compmake.plugins.backend_multyvac.sync import clean_cloud_out, sync_data_down, sync_data_up
from compmake.plugins.backend_multyvac.sync_db import delete_db_volume, synchronize_db_up
from compmake.ui import ACTIONS, ui_command
from compmake.ui.commands import raise_error_if_manager_failed
from .logging_imp import disable_logging_if_config

__all__ = ["cloudmake", "cloudclean", "cloud_sync_up", "cloud_sync_down"]


@ui_command(section=ACTIONS, dbchange=True)
def cloudmake(
    job_list,
    context,
    cq,
    n=DefaultsToConfig("multyvac_max_jobs"),
    recurse=DefaultsToConfig("recurse"),
    new_process=DefaultsToConfig("new_process"),
    echo=DefaultsToConfig("echo"),
    skipsync=False,
    rdb=True,
):
    """
        Multyvac backend

    """
    # TODO: check it exists
    # noinspection PyUnresolvedReferences
    import multyvac

    disable_logging_if_config(context)

    publish(context, "parmake-status", status="Obtaining job list")
    job_list = list(job_list)

    db = context.get_compmake_db()
    if not job_list:
        # XXX
        job_list = list(top_targets(db=db))

    volumes = sync_data_up(context, skipsync)

    if rdb:
        rdb_vol, rdb_db = synchronize_db_up(context, job_list)
    else:
        rdb_vol, rdb_db = None, None

    publish(context, "parmake-status", status="Starting multiprocessing manager (forking)")
    manager = MVacManager(
        num_processes=n,
        context=context,
        cq=cq,
        recurse=recurse,
        new_process=new_process,
        show_output=echo,
        volumes=volumes,
        rdb=rdb,
        rdb_vol=rdb_vol,
        rdb_db=rdb_db,
    )

    publish(context, "parmake-status", status="Adding %d targets." % len(job_list))
    manager.add_targets(job_list)

    publish(context, "parmake-status", status="Processing")
    manager.process()

    if not skipsync:
        sync_data_down(context)

    return raise_error_if_manager_failed(manager)


@ui_command(alias="cloud-clean", section=ACTIONS, dbchange=False)
def cloudclean(context):
    """ Cleans all jobs and results on the remote DB. """
    disable_logging_if_config(context)
    db = context.get_compmake_db()
    delete_db_volume(db)
    clean_cloud_out()


@ui_command(alias="cloud-sync-up", section=ACTIONS, dbchange=False)
def cloud_sync_up(context):
    """ Synchronizes local input data to the cloud. """
    disable_logging_if_config(context)
    sync_data_up(context)


@ui_command(alias="cloud-sync-down", section=ACTIONS, dbchange=False)
def cloud_sync_down(context):
    """ Synchronizes remote output data to local dir. """
    disable_logging_if_config(context)
    sync_data_down(context)



from collections import namedtuple
import sys

from compmake import CompmakeConstants
from .utils import AvgSystemStats


class CompmakeGlobalState:
    original_stderr = sys.stderr
    original_stdout = sys.stdout

    compmake_status = None

    class EventHandlers:
        # event name -> list of functions
        handlers = {}
        # list of handler, called when there is no other specialized handler
        fallback = []

    # TODO: make configurable
    system_stats = AvgSystemStats(interval=0.1, history_len=10)

    # Configuration vlues
    compmake_config = {}
    # config name -> ConfigSwitch
    config_switches = {}
    # section name -> ConfigSection
    config_sections = {}

    # Cached list of options for completions in console
    cached_completions = None


def get_compmake_config(key):
    config = CompmakeGlobalState.compmake_config
    if not key in config:
        msg = "Config %r not found. Known: %s" % (key, list(config))
        raise ValueError(msg)
    return config[key]


def set_compmake_config(key: str, value):
    # TODO: check exists
    CompmakeGlobalState.compmake_config[key] = value


ConfigSwitch = namedtuple("ConfigSwitch", "name default_value desc section order allowed")
ConfigSection = namedtuple("ConfigSection", "name desc order switches")


def set_compmake_status(s: str):
    CompmakeGlobalState.compmake_status = s


def is_interactive_session() -> bool:
    """ If this is true, we will ask questions to the user. """
    return get_compmake_status() == CompmakeConstants.compmake_status_interactive


def get_compmake_status() -> str:
    return CompmakeGlobalState.compmake_status
